# -*- coding: utf-8 -*-

"""Project Modules."""


from .modules import HookProxy, ProjectModule
from .packages import DevPackagesModule, PackagesModule
from .stubs import StubsModule
from .templates import TemplatesModule

__all__ = [
    "TemplatesModule",
    "PackagesModule",
    "StubsModule",
    "ProjectModule",
    "DevPackagesModule",
    "HookProxy",
]



"""Micropy stub utils."""

import importlib.util
import sys
from pathlib import Path
from types import ModuleType

import micropy.data


def import_source_code(module_name: str, path: Path) -> ModuleType:
    """Dynamically create and load module from python source code.

    Args:
        module_name: name of new module to create.
        path: path to source code.

    Returns:
        Dynamically created module.

    """
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def import_stubber() -> ModuleType:
    """Dynamically import stubber.

    We do this because `micropython-stubs` is not a python package, so
    we can't import from it as you would normally.

    """
    vers_path = micropy.data.STUBBER / "src" / "version.py"
    src_path = micropy.data.STUBBER / "src" / "utils.py"
    # stubber utils expects an ambiguous 'version' import
    import_source_code("version", vers_path)
    mod = import_source_code("stubber.utils", src_path)
    return mod


def generate_stub(path, log_func=None):
    """Create Stub from local .py file.

    Args:
        path (str): Path to file
        log_func (func, optional): Callback function for logging.
            Defaults to None.

    Returns:
        tuple: Tuple of file path and generated stub path.

    """
    stubgen = import_stubber()
    # Monkeypatch print to prevent or wrap output
    logfn = log_func or (lambda *a: None)
    stubgen.print = logfn
    stubgen.stubgen.print = logfn
    file_path = Path(path).absolute()
    stubbed_path = file_path.with_suffix(".pyi")
    stubgen.generate_pyi_from_file(file_path)  # noqa
    # ensure stubs reside next to their source.
    result = next((file_path.parent.rglob(f"**/{stubbed_path.name}")), stubbed_path)
    if result.exists():
        result.replace(stubbed_path)
    if not any(result.parent.iterdir()):
        result.parent.rmdir()
    files = (file_path, stubbed_path)
    return files



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""MicropyCli Console Entrypoint."""
import sys
from pathlib import Path

import click
import micropy.exceptions as exc
import questionary as prompt
from micropy import main, utils
from micropy.logger import Log
from micropy.project import Project, modules
from questionary import Choice

pass_mpy = click.make_pass_decorator(main.MicroPy, ensure=True)


@click.group(invoke_without_command=True)
@click.version_option()
@click.option(
    "--skip-checks",
    "-s",
    is_flag=True,
    default=False,
    help="Skip Project Checks. Defaults to False.",
)
@pass_mpy
@click.pass_context
def cli(ctx, mpy, skip_checks=False):
    """CLI Application for creating/managing Micropython Projects."""
    if ctx.invoked_subcommand is None:
        if not mpy.project.exists:
            return click.echo(ctx.get_help())
    latest = utils.is_update_available()
    if latest:
        log = Log.get_logger("MicroPy")
        log.title("Update Available!")
        log.info(f"Version $B[v{latest}] is now available")
        log.info("You can update via: $[pip install --upgrade micropy-cli]\n")
    mpy.RUN_CHECKS = not skip_checks


@cli.group(short_help="Manage Micropy Stubs")
def stubs():
    """Manage Micropy Stubs.

    \b
    Stub files are what enable linting,
    Intellisense, Autocompletion, and more.

    \b
    To achieve the best results, you can install
    stubs specific to your device/firmware using:

        micropy stubs add <STUB_NAME>

    For more info, please check micropy stubs add --help

    """


@cli.command(short_help="Create new Micropython Project")
@click.argument("path", required=False, default=None)
@click.option(
    "--name", "-n", required=False, default=None, help="Project Name. Defaults to Path name."
)
@click.option(
    "--template",
    "-t",
    type=click.Choice(modules.TemplatesModule.TEMPLATES.keys()),
    multiple=True,
    required=False,
    help=("Templates to generate for project." " Multiple options can be passed."),
)
@pass_mpy
def init(mpy, path, name=None, template=None):
    """Create new Micropython Project.

    \b When creating a new project, all files will be placed under the
    generated <PROJECT_NAME> folder.

    """
    mpy.log.title("Creating New Project")
    if not path:
        path = Path.cwd()
        default_name = path.name
        prompt_name = prompt.text("Project Name", default=default_name).ask()
        name = prompt_name.strip()
    if not template:
        templates = modules.TemplatesModule.TEMPLATES.items()
        templ_choices = [Choice(str(val[1]), value=t) for t, val in templates]
        template = prompt.checkbox(f"Choose any Templates to Generate", choices=templ_choices).ask()
    stubs = [Choice(str(s), value=s) for s in mpy.stubs]
    if not stubs:
        mpy.log.error("You don't have any stubs!")
        mpy.log.title("To add stubs to micropy, use $[micropy stubs add <STUB_NAME>]")
        sys.exit(1)
    stub_choices = prompt.checkbox(f"Which stubs would you like to use?", choices=stubs).ask()
    project = Project(path, name=name)
    project.add(modules.StubsModule, mpy.stubs, stubs=stub_choices)
    project.add(modules.PackagesModule, "requirements.txt")
    project.add(modules.DevPackagesModule, "dev-requirements.txt")
    project.add(modules.TemplatesModule, templates=template, run_checks=mpy.RUN_CHECKS)
    proj_path = project.create()
    try:
        rel_path = f"./{proj_path.relative_to(Path.cwd())}"
    except ValueError:
        rel_path = proj_path
    mpy.log.title(f"Created $w[{project.name}] at $w[{rel_path}]")


@cli.command(short_help="Install Project Requirements")
@click.argument("packages", nargs=-1)
@click.option(
    "-d", "--dev", is_flag=True, default=False, help=("Add dependency to dev requirements")
)
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True, path_type=str),
    help=("Add dependency from local path. Can be a file or directory."),
)
@pass_mpy
def install(mpy, packages, dev=False, path=None):
    """Install Packages as Project Requirements.

    \b
    Install a project dependency while enabling
    intellisense, autocompletion, and linting for it.

    \b
    If no packages are passed and a requirements.txt file is found,
    then micropy will install all packages listed in it.

    \b
    If the --dev flag is passed, then the packages are only
    added to micropy.json. They are not stubbed.

    \b
    To add a dependency from a path, use the --path option
    and provide a name for your package:
        \b
        $ micropy install --path ./src/lib/mypackage MyCustomPackage


    \b
    You can import installed packages just as you would
    on your actual device:
        \b
        >>> # main.py
        >>> import <package_name>

    """
    project = mpy.project
    if not project.exists:
        mpy.log.error("You are not currently in an active project!")
        sys.exit(1)
    if path:
        pkg_name = next(iter(packages), None)
        mpy.log.title("Installing Local Package")
        pkg_path = "-e " + path
        project.add_package(pkg_path, dev=dev, name=pkg_name)
        sys.exit(0)
    if not packages:
        mpy.log.title("Installing all Requirements")
        try:
            project.add_from_file(dev=dev)
        except Exception as e:
            mpy.error(f"Failed to load requirements!", exception=e)
            raise click.Abort()
        else:
            mpy.log.success("\nRequirements Installed!")
            sys.exit(0)
    mpy.log.title("Installing Packages")
    for pkg in packages:
        try:
            project.add_package(pkg, dev=dev)
        except exc.RequirementException as e:
            pkg_name = str(e.package)
            mpy.log.error(
                (f"Failed to install {pkg_name}!" " Is it available on PyPi?"), exception=e
            )
            raise click.Abort()


@stubs.command(short_help="Add Stubs from package or path")
@click.argument("stub_name", required=True)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite Stub if it exists.")
@pass_mpy
def add(mpy, stub_name, force=False):
    """Add Stubs from package or path.

    \b
    In general, stub package names follow this schema:
        <device>-<firmware>-<version>

    \b
    For example:
        esp32-micropython-1.11.0

    \b
    You can search premade stub packages using:
        micropy stubs search <QUERY>

    Checkout the docs on Github for more info.

    """
    mpy.stubs.verbose_log(True)
    proj = mpy.project
    mpy.log.title(f"Adding $[{stub_name}] to stubs")
    try:
        stub = mpy.stubs.add(stub_name, force=force)
    except exc.StubNotFound:
        mpy.log.error(f"$[{stub_name}] could not be found!")
        sys.exit(1)
    except exc.StubError:
        mpy.log.error(f"$[{stub_name}] is not a valid stub!")
        sys.exit(1)
    else:
        if proj.exists:
            mpy.log.title(f"Adding $[{stub.name}] to $[{proj.name}]")
            proj.add_stub(stub)


@stubs.command()
@click.argument("query", required=True)
@pass_mpy
def search(mpy, query):
    """Search available Stubs."""
    mpy.log.title(f"Searching Stub Repositories...")
    results = mpy.stubs.search_remote(query)
    mpy.log.title(f"Results for $[{query}]:")
    for pkg, installed in results:
        name = f"{pkg} $B[(Installed)]" if installed else pkg
        mpy.log.info(name)


@stubs.command()
@pass_mpy
def list(mpy):
    """List installed stubs."""

    def print_stubs(stub_list):
        for firm, stubs in stub_list:
            if stubs:
                title = str(firm).capitalize()
                mpy.log.title(f"$[{title}]:")
                for stub in stubs:
                    mpy.log.info(str(stub))

    mpy.log.title("Installed Stubs:")
    mpy.log.info(f"Total: {len(mpy.stubs)}")
    print_stubs(mpy.stubs.iter_by_firmware())
    mpy.verbose = False
    proj = mpy.project
    if proj.exists:
        mpy.log.title(f"Stubs used in {proj.name}:")
        mpy.log.info(f"Total: {len(proj.stubs)}")
        stubs = mpy.stubs.iter_by_firmware(stubs=proj.stubs)
        print_stubs(stubs)


@stubs.command(short_help="Create Stubs from Pyboard")
@click.argument("port", required=True)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Enable verbose output")
@pass_mpy
def create(mpy, port, verbose=False):
    """Create stubs from a pyboard at <PORT>

    \b
    MicropyCli uses Josverl's micropython-stubber for stub creation.
    For more information, please visit the repository
    at: https://github.com/Josverl/micropython-stubber

    """
    return mpy.create_stubs(port, verbose=verbose)


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-

"""Project Packages Module."""

import shutil
from pathlib import Path
from typing import Any, Optional, Union

from boltons import fileutils
from micropy import utils
from micropy.config import Config
from micropy.packages import (
    LocalDependencySource,
    Package,
    PackageDependencySource,
    create_dependency_source,
)
from micropy.project.modules import ProjectModule


class PackagesModule(ProjectModule):
    """Project Module for handling requirements.

    Args:
        path (str): Path to create requirements file at.
        packages (dict, optional): Initial packages to use.
            Defaults to None.

    """

    name: str = "packages"
    PRIORITY: int = 7

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self._path = Path(path)

    @property
    def packages(self):
        _packages = self.config.get(self.name, {})
        return _packages

    @property
    def path(self):
        """Path to requirements file.

        Returns:
            Path: Path to file

        """
        path = self.parent.path / self._path
        return path

    @property
    def pkg_path(self):
        """Path to package data folder.

        Returns:
            Path: Path to folder.

        """
        return self.parent.data_path / self.parent.name

    @property
    def config(self) -> Config:
        """Config values specific to component.

        Returns:
            Component config.

        """
        return self.parent.config

    @property
    def context(self) -> Config:
        """Context values specific to component.

        Returns:
            Context values.

        """
        return self.parent.context

    @property
    def cache(self) -> Config:
        """Project Cache.

        Returns:
            Project wide cache

        """
        return self.parent.cache

    def install_package(self, source: Union[LocalDependencySource, PackageDependencySource]) -> Any:
        with source as files:
            if source.is_local:
                self.log.debug(f"installing {source} as local")
                return
            if isinstance(files, list):
                self.log.debug(f"installing {source} as module(s)")
                # Iterates over flattened list of stubs tuple
                file_paths = [(f, (self.pkg_path / f.name)) for f in list(sum(files, ()))]
                for paths in file_paths:
                    shutil.move(*paths)  # overwrites if existing
                return file_paths
            self.log.debug(f"installing {source} as package")
            pkg_path = self.pkg_path / source.package.name
            return fileutils.copytree(files, pkg_path)

    @ProjectModule.hook(dev=False)
    def add_from_file(self, path: Optional[Path] = None, dev: bool = False, **kwargs: Any) -> dict:
        """Loads all requirements from file.

        Args:
            path: Path to file. Defaults to self.path.
            dev: If dev requirements should be loaded.
                Defaults to False.

        """
        path = path or self.path
        reqs = utils.iter_requirements(path)
        self.log.debug(f"loading requirements from: {path}")
        for r in reqs:
            pkg = create_dependency_source(r.line).package
            if not self.packages.get(pkg.name):
                self.config.add(self.name + "/" + pkg.name, pkg.pretty_specs)
                if pkg.editable:
                    self.context.extend("local_paths", [pkg.path], unique=True)
        return self.packages

    @ProjectModule.hook()
    def add_package(self, package, dev=False, **kwargs):
        """Add requirement to project.

        Args:
            package (str): package name/spec
            dev (bool, optional): If dev requirements should be loaded.
                Defaults to False.

        Returns:
            dict: Dictionary of packages

        """
        self.log.debug(f"adding new dependency: {package}")
        source = create_dependency_source(package, **kwargs)
        pkg = source.package
        self.log.info(f"Adding $[{pkg.name}] to requirements...")
        if self.packages.get(pkg.name, None):
            self.log.error(f"$[{pkg}] is already installed!")
            self.update()
            return None
        self.config.add(self.name + "/" + pkg.name, pkg.pretty_specs)
        try:
            self.load()
        except Exception as e:
            self.log.error(f"Failed to install: {pkg.name}", exception=e)
            self.config.pop(self.name + "/" + pkg.name)
            raise
        else:
            if pkg.editable:
                self.context.extend("local_paths", [pkg.path], unique=True)
            self.log.success("Package installed!")
        finally:
            self.parent.update()
            return self.packages

    def load(self, fetch=True, **kwargs):
        """Retrieves and stubs project requirements."""
        self.pkg_path.mkdir(parents=True, exist_ok=True)
        if self.path.exists():
            self.add_from_file(self.path)
        pkg_keys = set(self.packages.keys())
        pkg_cache = self.cache.get(self.name)
        new_pkgs = pkg_keys.copy()
        if pkg_cache:
            new_pkgs = new_pkgs - set(pkg_cache)
        new_packages = [
            Package.from_text(name, spec)
            for name, spec in self.packages.items()
            if name in new_pkgs
        ]
        if fetch:
            if new_packages:
                self.log.title("Fetching Requirements")
            for req in new_packages:

                def format_desc(p):
                    return "".join(self.log.iter_formatted(f"$B[{p}]"))

                source = create_dependency_source(
                    str(req),
                    name=req.name,
                    format_desc=lambda p: f"{self.log.get_service()} {format_desc(p)}",
                )
                self.install_package(source)
        self.update()
        self.cache.upsert(self.name, list(pkg_keys))

    def create(self):
        """Create project files."""
        self.pkg_path.mkdir(parents=True, exist_ok=True)
        if not self.config.get(self.name):
            self.config.add(self.name, {})
        return self.update()

    def update(self):
        """Dumps packages to file at path."""
        if not self.path.exists():
            self.path.touch()
        pkgs = [Package.from_text(name, spec) for name, spec in self.config.get(self.name).items()]
        self.log.debug(f"dumping to {self.path.name}")
        with self.path.open("r+") as f:
            content = [c.strip() for c in f.readlines() if c.strip() != ""]
            _lines = sorted(set(str(p) for p in pkgs) | set(content))
            lines = [l + "\n" for l in _lines]
            self.log.debug(f"dumping: {lines}")
            f.seek(0)
            f.writelines(lines)
        local_paths = [p.path for p in pkgs if p.editable]
        if local_paths:
            self.context.add("local_paths", local_paths)
        self.context.extend("paths", [self.pkg_path], unique=True)


class DevPackagesModule(PackagesModule):
    """Project Module for Dev Packages."""

    PRIORITY: int = 8

    def __init__(self, path, **kwargs):
        super().__init__(path, **kwargs)
        self.name = "dev-packages"

    def create(self):
        """Creates component."""
        self.config.add(f"{self.name}/micropy-cli", "*")
        super().create()

    def load(self, *args, **kwargs):
        """Load component."""
        super().load(*args, **kwargs, fetch=False)

    @ProjectModule.hook(dev=True)
    def add_package(self, package, **kwargs):
        """Adds package."""
        return super().add_package(package, **kwargs)

    @ProjectModule.hook(dev=True)
    def add_from_file(self, path=None, **kwargs):
        """Adds packages from file."""
        return super().add_from_file(path=path, **kwargs)



# -*- coding: utf-8 -*-

"""Hosts functionality relating to generation of user projects."""

from pathlib import Path
from typing import Any, Iterator, List, Optional, Type

from boltons.queueutils import PriorityQueue
from micropy.config import Config, DictConfigSource
from micropy.logger import Log, ServiceLog
from micropy.project.modules import ProjectModule


class Project(ProjectModule):
    """Micropy Project.

    Args:
        path (str): Path to project root.
        name (str, optional): Name of Project.
            Defaults to None. If none, uses name of current directory.

    """

    def __init__(self, path: str, name: Optional[str] = None, **kwargs: Any):
        self._children: List[Type[ProjectModule]] = []
        self.path: Path = Path(path).absolute()
        self.data_path: Path = self.path / ".micropy"
        self.info_path: Path = self.path / "micropy.json"
        self.cache_path: Path = self.data_path / ".cache"
        self._cache = Config(self.cache_path)
        self._context = Config(source_format=DictConfigSource, default={"datadir": self.data_path})
        self.name: str = name or self.path.name
        default_config = {
            "name": self.name,
        }
        self._config: Config = Config(self.info_path, default=default_config)
        self.log: ServiceLog = Log.add_logger(self.name, show_title=False)

    def __getattr__(self, name: str) -> Any:
        results = iter([c.resolve_hook(name) for c in self._children])
        for res in results:
            if res is not None:
                self.log.debug(f"Hook Resolved: {name} -> {res}")
                return res
        return self.__getattribute__(name)

    @property
    def exists(self) -> bool:
        """Whether this project exists.

        Returns:
            bool: True if it exists

        """
        return self.info_path.exists()

    @property
    def config(self) -> Config:
        """Project Configuration.

        Returns:
            Config: Project Config Instance

        """
        return self._config

    @property
    def context(self) -> Config:
        """Project context used in templates.

        Returns:
            Config: Current context

        """
        return self._context

    @property
    def cache(self) -> Config:
        """Project wide cache.

        Returns:
            Cache instance

        """
        return self._cache

    def iter_children_by_priority(self) -> Iterator[Type[ProjectModule]]:
        """Iterate project modules by priority.

        Yields:
            the next child item

        """
        pq = PriorityQueue()
        for i in self._children:
            pq.add(i, i.PRIORITY)
        more = pq.peek(default=False)
        while more:
            yield pq.pop()
            more = pq.peek(default=False)

    def add(self, component, *args, **kwargs):
        """Adds project component.

        Args:
            component (Any): Component to add.

        """
        child = component(*args, **kwargs, log=self.log, parent=self)
        self._children.append(child)
        self.log.debug(f"adding module: {type(child).__name__}")

    def remove(self, component):
        """Removes project component.

        Args:
            component (Any): Component to remove.

        """
        child = next(i for i in self._children if isinstance(i, component))
        self._children.remove(child)

    def load(self, **kwargs: Any) -> "Project":
        """Loads all components in Project.

        Returns:
            Current Project Instance

        """
        self.name = self._config.get("name")
        self.data_path.mkdir(exist_ok=True)
        for child in self.iter_children_by_priority():
            child.load(**kwargs)
        return self

    def create(self):
        """Creates new Project.

        Returns:
            Path: Path relative to current active directory.

        """
        self.log.title(f"Initiating $[{self.name}]")
        self.data_path.mkdir(exist_ok=True, parents=True)
        ignore_data = self.data_path / ".gitignore"
        ignore_data.write_text("*")
        self.log.debug(f"Generated Project Context: {self.context}")
        for child in self.iter_children_by_priority():
            child.create()
        self.info_path.touch()
        self.config.sync()
        self.log.success(f"Project Created!")
        return self.path

    def update(self):
        """Updates all project components.

        Returns:
            Current active project.

        """
        self.log.debug("Updating all project modules...")
        for child in self.iter_children_by_priority():
            child.update()
        return self



try:
    import usocket as socket
except:
    import socket
try:
    import ustruct as struct
except:
    import struct

# (date(2000, 1, 1) - date(1900, 1, 1)).days * 24*60*60
NTP_DELTA = 3155673600

host = "pool.ntp.org"


def time():
    NTP_QUERY = bytearray(48)
    NTP_QUERY[0] = 0x1B
    addr = socket.getaddrinfo(host, 123)[0][-1]
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(1)
    res = s.sendto(NTP_QUERY, addr)
    msg = s.recv(48)
    s.close()
    val = struct.unpack("!I", msg[40:44])[0]
    return val - NTP_DELTA


# There's currently no timezone support in MicroPython, so
# utime.localtime() will return UTC time (as if it was .gmtime())
def settime():
    t = time()
    import machine
    import utime

    tm = utime.localtime(t)
    tm = tm[0:3] + (0,) + tm[3:6] + (0,)
    machine.RTC().datetime(tm)
    print(utime.localtime())



# -*- coding: utf-8 -*-
""" Common Pytest Fixtures"""

import importlib
import json
import shutil
from pathlib import Path
from pprint import PrettyPrinter

import micropy
import pytest
import questionary
from boltons import iterutils

# Mock values for Template VSCode ext checks
mock_vscode_exts = [
    "mock.ext@0.0.0",
    # meets req
    "ms-python.python@2019.9.34474",
]


def pytest_collection_modifyitems(items):
    items.reverse()


@pytest.fixture(autouse=True)
def cleanup_data(mocker):
    mocker.resetall()
    try:
        micropy.utils.ensure_valid_url.clear_cache()
        micropy.stubs.source.StubRepo.repos = set()
    except Exception:
        importlib.reload(micropy)


@pytest.fixture
def mock_prompt(monkeypatch):
    def mock_prompt(*args, **kwargs):
        class prompt_mock:
            def __init__(self, *args, **kwargs):
                return None

            def ask(self):
                return ["stub"]

        return prompt_mock(*args, **kwargs)

    monkeypatch.setattr(questionary, "checkbox", mock_prompt)


@pytest.fixture
def mock_micropy_path(mocker, tmp_path):
    path = tmp_path / ".micropy"
    stub_path = path / "stubs"
    log_path = path / "micropy.log"
    mocker.patch("micropy.data.FILES", path)
    mocker.patch("micropy.data.STUB_DIR", stub_path)
    mocker.patch("micropy.data.LOG_FILE", log_path)
    return path


@pytest.fixture
def mock_micropy(mock_micropy_path):
    mp = micropy.main.MicroPy()
    return mp


@pytest.fixture
def mock_cwd(request, tmp_path, mocker):
    print(request)
    import pathlib

    mocker.patch("pathlib.Path.cwd")
    pathlib.Path.cwd.return_value = tmp_path
    yield (tmp_path)


@pytest.fixture(scope="session")
def test_urls():
    def test_headers(type):
        return {"content-type": type}

    return {
        "valid": "http://www.google.com",
        "valid_https": "https://www.google.com",
        "invalid": "/foobar/bar/foo",
        "invalid_file": "file:///foobar/bar/foo",
        "bad_resp": "http://www.google.com/XYZ/ABC/BADRESP",
        "download": "https://www.somewebsite.com/archive_test_stub.tar.gz",
        "headers": {
            "can_download": test_headers("application/gzip"),
            "not_download": test_headers("text/plain"),
        },
    }


@pytest.fixture
def get_stub_paths(shared_datadir, tmp_path):
    def _get_stub_paths(count=1, valid=True, firm=False, dest=tmp_path):
        _stubs = ["fware"] if firm else ["esp8266", "esp32"]
        stubs = iter(_stubs)
        _count = 0
        while _count < count:
            s = next(stubs)
            path = (
                (shared_datadir / f"{s}_test_stub")
                if valid
                else (shared_datadir / f"{s}_invalid_stub")
            )
            if path.exists():
                dest = dest / path.name
                if not dest.exists():
                    shutil.copytree(path, (dest / path.name))
                yield dest
                _count += 1

    return _get_stub_paths


@pytest.fixture
def mock_mp_stubs(mock_micropy, mocker, shared_datadir):
    mock_micropy.stubs.add((shared_datadir / "fware_test_stub"))
    mock_micropy.stubs.add((shared_datadir / "esp8266_test_stub"))
    mock_micropy.stubs.add((shared_datadir / "esp32_test_stub"))
    return mock_micropy


@pytest.fixture
def get_stubs(get_stub_paths, mocker, tmp_path):
    def _get_stubs(path=tmp_path, **kwargs):
        def stubbify(m, path, firm=None):
            m.path = path
            m.frozen = path / "frozen"
            m.stubs = path / "stubs"
            m.name = m.path.name
            m.stub_version = "0.0.0"
            m.firmware = firm
            return m

        paths = get_stub_paths(dest=path, **kwargs)
        firm = next(get_stub_paths(firm=True, dest=path))
        firm_mock = stubbify(mocker.MagicMock(), firm)
        for p in paths:
            yield stubbify(mocker.MagicMock(), p, firm=firm_mock)

    return _get_stubs


@pytest.fixture
def micropy_stubs(mocker, get_stubs):
    def _micropy_stubs(count=3):
        def _mock_resolve_subresource(stubs, data_path):
            return get_stubs(path=data_path)

        mock_mp = mocker.patch.object(micropy, "MicroPy").return_value
        stubs = list(get_stubs())
        mock_mp.stubs.__iter__.return_value = stubs
        mock_mp.stubs.resolve_subresource = _mock_resolve_subresource
        mock_mp.stubs.add.return_value = stubs[0]
        return mock_mp

    return _micropy_stubs


@pytest.yield_fixture
def test_archive(shared_datadir):
    archive = shared_datadir / "archive_test_stub.tar.gz"
    file_obj = archive.open("rb")
    file_bytes = file_obj.read()
    yield file_bytes
    file_obj.close()


@pytest.fixture
def test_repo(test_urls, shared_datadir, mocker):
    micropy.stubs.source.StubRepo.repos = set()
    mocker.patch.object(micropy.stubs.source.utils, "is_downloadable", return_value=True)
    mocker.patch.object(
        micropy.stubs.source.utils, "ensure_valid_url", return_value=test_urls["valid"]
    )
    test_data = json.loads((shared_datadir / "test_repo.json").read_text())
    mock_get = mocker.patch.object(micropy.stubs.source.requests, "get")
    mock_get.return_value.json.return_value = test_repo
    repo = micropy.stubs.source.StubRepo(**test_data)
    return repo


@pytest.fixture
def mock_checks(mocker):
    """Mock VSCode Template Checks"""
    m_run = mocker.patch.object(micropy.project.checks.subproc, "run").return_value
    type(m_run).stdout = mocker.PropertyMock(return_value="\n".join(mock_vscode_exts))
    return m_run


@pytest.fixture
def mock_pkg(mocker, tmp_path):
    """return mock package"""
    from micropy import packages

    tmp_pkg = tmp_path / "tmp_pkg"
    tmp_pkg.mkdir()
    (tmp_pkg / "module.py").touch()
    (tmp_pkg / "file.py").touch()
    mocker.patch.object(packages.source_package.utils, "ensure_valid_url")
    mock_tarbytes = mocker.patch.object(packages.source_package.utils, "extract_tarbytes")
    mock_meta = mocker.patch.object(packages.source_package.utils, "get_package_meta")
    mocker.patch.object(packages.source_package.utils, "get_url_filename")
    mocker.patch.object(packages.source_package.utils, "stream_download")
    mock_tarbytes.return_value = tmp_pkg
    mock_meta.return_value = {"url": "http://realurl.com"}
    return tmp_pkg


# Pytest Incremental Marker
def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed ({})".format(previousfailed.name))


class AssertUtils:
    pp = PrettyPrinter(indent=4, width=20).pprint

    def dict_match_mocks(self, d1):
        from unittest.mock import Mock

        # Mocks
        remapped = iterutils.remap(
            d1, lambda p, k, v: (k, "MOCKED_VALUE") if isinstance(v, Mock) else True
        )
        return remapped

    def dict_equal(self, d1, d2):
        match_d1 = sorted(self.dict_match_mocks(d1).items())
        match_d2 = sorted(self.dict_match_mocks(d1).items())
        print("== IS DICT EQUAL ==")
        self.pp(match_d1)
        print("\n----------\n")
        self.pp(match_d2)
        print("==============")
        return match_d1 == match_d2

    def list_equal(self, l1, l2):
        list_one = sorted(l1)
        list_two = sorted(l2)
        return list_one == list_two

    def load_json(self, path):
        import json

        return json.loads(path.read_text())

    def json_equal_dict(self, path, d2):
        data = self.load_json(path)
        return self.dict_equal(data, d2)

    def str_path(self, path, absolute=False):
        """x-platform path strings helper"""
        path = Path(path)
        if absolute:
            path = path.absolute()
        return str(path)

    def get_rand_str(self, length=10):
        import random
        import string

        letters = string.ascii_lowercase
        return "".join(random.choice(letters) for i in range(length))


@pytest.fixture
def utils():
    return AssertUtils()



# -*- coding: utf-8 -*-

"""Various requirement checks for templates."""

import subprocess as subproc
from functools import partial as _p

from micropy.logger import Log
from packaging import version

VSCODE_MS_PY_MINVER = "2019.9.34474"

log = Log.get_logger("MicroPy")


def iter_vscode_ext(name=None):
    """Iterates over installed VSCode Extensions.

    Args:
        name (str, optional): Name of Extension to Yield

    """
    _cmd = "code --list-extensions --show-versions"
    proc = subproc.run(_cmd, stdout=subproc.PIPE, stderr=subproc.PIPE, shell=True)
    results = [e.strip() for e in proc.stdout.splitlines()]
    for ext in results:
        ename, vers = ext.split("@")
        if not name:
            yield (ename, vers)
        if name and ename == name:
            yield (ename, vers)


def vscode_ext_min_version(ext, min_version=VSCODE_MS_PY_MINVER, info=None):
    """Check if installed VScode Extension meets requirements.

    Args:
        ext (str): Name of Extension to Test
        min_version (str, optional): Minimum version.
            Defaults to VSCODE_MS_PY_MINVER.
        info (str, optional): Additional information to output.
            Defaults to None.

    Returns:
        bool: True if requirement is satisfied, False otherwise.

    """
    try:
        name, vers = next(iter_vscode_ext(name=ext), (ext, "0.0.0"))
    except Exception as e:
        log.debug(f"vscode check failed to run: {e}")
        log.debug("skipping...")
        return True
    else:
        cur_vers = version.parse(vers)
        min_vers = version.parse(min_version)
        if cur_vers >= min_vers:
            return True
        log.error(f"\nVSCode Extension {ext} failed to satisfy requirements!", bold=True)
        log.error(f"$[Min Required Version]: {min_vers}")
        log.error(f"$[Current Version:] {cur_vers}")
        if info:
            log.warn(info)
        return False


TEMPLATE_CHECKS = {
    "ms-python": _p(
        vscode_ext_min_version,
        "ms-python.python",
        info=("VSCode Integration will fail! " "See $[BradenM/micropy-cli#50] for details.\n"),
    ),
}



#!/usr/bin/env python3
"""
Collect modules and python stubs from the Loboris MicroPython source project and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder
"""

# pylint: disable= line-too-long
# Copyright (c) 2020 Jos Verlinde
# MIT license
import logging
from pathlib import Path
import downloader
import utils

FAMILY = "loboris"
PORT = "esp32_lobo"


log = logging.getLogger(__name__)
# log.setLevel(level=logging.DEBUG)


def get_frozen(stub_path=None, *, repo=None, version="3.2.24"):
    "Loboris frozen modules"
    if stub_path is None:
        stub_path = Path("./all-stubs") / "{}-{}-frozen".format(FAMILY, utils.flat_version(version))
    else:
        stub_path = Path(stub_path)

    if not repo:
        repo = "https://raw.githubusercontent.com/loboris/MicroPython_ESP32_psRAM_LoBo/master/MicroPython_BUILD/components/micropython/esp32/modules/{}"

    frozen_modules = [
        "README.md",
        "ak8963.py",
        "freesans20.py",
        "functools.py",
        "logging.py",
        "microWebSocket.py",
        "microWebSrv.py",
        "microWebTemplate.py",
        "mpu6500.py",
        "mpu9250.py",
        "pye.py",
        "ssd1306.py",
        "tpcalib.py",
        "upip.py",
        "upip_utarfile.py",
        "upysh.py",
        "urequests.py",
        "writer.py",
    ]
    # download
    downloader.download_files(repo, frozen_modules, stub_path)
    # make a manifest
    utils.make_manifest(stub_path, FAMILY, port="esp32", version=version)


if __name__ == "__main__":
    logging.basicConfig(format="%(levelname)-8s:%(message)s", level=logging.INFO)
    get_frozen(version="3.2.24")



#!/usr/bin/env python3
"""
Collect modules and python stubs from MicroPython source projects (v1.12 +) and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder

"""
# pylint: disable= line-too-long,  W1202, invalid-name

# Copyright (c) 2020 Jos Verlinde
# MIT license
# some functions used from micropython/micropython/tools/makemanifest.py,
#   part of the MicroPython project, http://micropython.org/
#   Copyright (c) 2019 Damien P. George

# locating frozen modules :
# tested on MicroPython v1.12 - v1.13
# - 1.13 - using manifests.py, and support for variant
# - 1.12 - using manifests.py, possible also include content of /port/modules folder ?
# - 1.11 and older - include content of /port/modules folder if it exists
import os
import glob
import re
import shutil
import warnings
import logging
import basicgit as git
import utils
from pathlib import Path  # start moving from os & glob to pathlib

log = logging.getLogger(__name__)
# log.setLevel(level=logging.DEBUG)

# globals
FAMILY = "micropython"

path_vars = {"MPY_DIR": "", "MPY_LIB_DIR": "", "PORT_DIR": "", "BOARD_DIR": ""}
stub_dir = None

# functions from makemanifest to ensure that the manifest.py files can be processed
class FreezeError(Exception):
    pass


# do not change method name
# freeze_as_mpy is only used by the unix port.
def freeze_as_mpy(path, script=None, opt=0):
    log.debug(" - freeze_as_mpy({},script={},opt={})".format(path, script, opt))
    freeze(path, script, opt)


# do not change method name
# TODO: see if/how `freeze_as_str` should be handled, may need re-implementing based on the v1.13 code
def freeze_as_str(path):
    log.debug(" - freeze_as_str({})".format(path))
    log.warning("WARNING: Currently freeze_as_string({:s}) is not supported/processed".format(path))


# def freeze_as_str(path):
#     """Freeze the given `path` and all .py scripts within it as a string,
#     which will be compiled upon import.
#     """
#     freeze_internal(KIND_AS_STR, path, None, 0)


# function used commonly in manifest.py to freeze a set of scripts
# pylint: disable=unused-argument, unused-variable
# do not change method name
def freeze(path, script=None, opt=0):
    """Freeze the input, automatically determining its type.  A .py script
    will be compiled to a .mpy first then frozen, and a .mpy file will be
    frozen directly.

    `path` must be a directory, which is the base directory to search for
    files from.  When importing the resulting frozen modules, the name of
    the module will start after `path`, ie `path` is excluded from the
    module name.

    If `path` is relative, it is resolved to the current manifest.py.
    Use $(MPY_DIR), $(MPY_LIB_DIR), $(PORT_DIR), $(BOARD_DIR) if you need
    to access specific paths.

    If `script` is None all files in `path` will be frozen.

    If `script` is an iterable then freeze() is called on all items of the
    iterable (with the same `path` and `opt` passed through).

    If `script` is a string then it specifies the filename to freeze, and
    can include extra directories before the file.  The file will be
    searched for in `path`.

    `opt` is the optimisation level to pass to mpy-cross when compiling .py
    to .mpy. (ignored in this implementation)
    """
    log.debug(" - freeze(({},script={},opt={})".format(path, script, opt))
    path = convert_path(path)
    if script is None:
        # folder of scripts.
        # for s in os.listdir(path):
        #     freezedry(path, s)

        for dirpath, dirnames, filenames in os.walk(path, followlinks=True):
            for script in filenames:
                # can recurse folder, so add relative path to script.
                freezedry(path, (dirpath + "/" + script)[len(path) + 1 :])
                # freeze_internal(kind, path, (dirpath + '/' + f)[len(path) + 1:], opt)
    elif not isinstance(script, str):
        # several specific scripts.
        for script in script:
            freezedry(path, script)
    else:
        # on specific script, may include a path: 'umqtt/simple.py'
        freezedry(path, script)


# called by freeze.
# do not change method name
def freezedry(path, script):
    "copy the to-be-frozen module to the destination folder to be stubbed"

    log.debug(" - freezedry({},{})".format(path, script))

    script_path = os.path.join(path, script)

    if stub_dir:
        log.info("freezedry : {:<30} to {}".format(script, stub_dir))
        dest_path = os.path.dirname(os.path.join(stub_dir, script))
        # ensure folder, including possible path prefix for script
        os.makedirs(dest_path, exist_ok=True)
        # copy file
        try:
            shutil.copy2(script_path, dest_path)
        except OSError as e:
            log.exception(e)
    else:
        log.error("Stub folder not set")


# do not change method name
def include(manifest):
    """
    Include another manifest.

    The manifest argument can be a string (filename) or an iterable of
    strings.

    Relative paths are resolved with respect to the current manifest file.
    """
    if not isinstance(manifest, str):
        for m in manifest:
            include(m)
    else:
        manifest = convert_path(manifest)
        with open(manifest) as f:
            # Make paths relative to this manifest file while processing it.
            # Applies to includes and input files.
            prev_cwd = os.getcwd()
            os.chdir(os.path.dirname(manifest))
            try:
                exec(f.read())  # pylint: disable=exec-used
            except OSError:
                log.warning("Could not process manifest: {}".format(manifest))
            os.chdir(prev_cwd)


def convert_path(path):
    "Perform variable substitution in path"
    for name, value in path_vars.items():
        path = path.replace("$({})".format(name), value)
    # Convert to absolute path (so that future operations don't rely on
    # still being chdir 'ed).
    return os.path.abspath(path)


def get_frozen(
    stub_path: str,
    version: str,
    mpy_path: str = None,
    lib_path: str = None,
):
    """
    get and parse the to-be-frozen .py modules for micropython to extract the static type information
     - requires that the MicroPython and Micropython-lib repos are checked out and available on a local path
     - repos should be cloned side-by-side as some of the manifests refer to micropython-lib scripts using a relative path
    """

    if not mpy_path:
        mpy_path = "./micropython"
    if not lib_path:
        lib_path = "./micropython-lib"
    if not stub_path:
        stub_path = "{}/{}_{}_frozen".format(utils.STUB_FOLDER, FAMILY, utils.flat_version(version))
    # get the manifests of the different ports and boards
    mpy_path = os.path.abspath(mpy_path)
    lib_path = os.path.abspath(lib_path)
    stub_path = os.path.abspath(stub_path)

    # manifest.py is used for board specific and daily builds
    # manifest_release.py is used for the release builds
    manifests = glob.glob(mpy_path + "/ports/**/manifest.py", recursive=True) + glob.glob(
        mpy_path + "/ports/**/manifest_release.py", recursive=True
    )

    # remove any manifests  that are below one of the virtual environments (venv) \
    # 'C:\\develop\\MyPython\\micropython\\ports\\esp32\\build-venv\\lib64\\python3.6\\site-packages\\pip\\_vendor\\distlib\\manifest.py'
    manifests = [m for m in manifests if not "venv" in m]

    if len(manifests) > 0:
        log.debug("MicroPython v1.12 and newer")
        get_frozen_manifest(manifests, stub_path, mpy_path, lib_path, version)
    else:
        log.debug("MicroPython v1.11, older or other")
        # others
        get_frozen_folders(stub_path, mpy_path, lib_path, version)


def get_frozen_folders(stub_path: str, mpy_path: str, lib_path: str, version: str):
    """
    get and parse the to-be-frozen .py modules for micropython to extract the static type information
    locates the to-be-frozen files in modules folders
    - 'ports/<port>/modules/*.py'
    - 'ports/<port>/boards/<board>/modules/*.py'
    """
    targets = []
    scripts = glob.glob(mpy_path + "/ports/**/modules/*.py", recursive=True)
    for script in scripts:
        mpy_port, mpy_board = get_target_names(script)
        if not mpy_board:
            mpy_board = "GENERIC"

        dest_path = os.path.join(stub_path, mpy_port, mpy_board)
        log.info("freezedry : {:<30} to {}".format(script, dest_path))
        # ensure folder, including possible path prefix for script
        os.makedirs(dest_path, exist_ok=True)
        # copy file
        try:
            shutil.copy2(script, dest_path)
            if not dest_path in targets:
                targets.append(dest_path)
        except OSError as e:
            ## Ignore errors that are caused by reorganisation of Micropython-lib
            # print(e)
            warnings.warn("unable to freeze {} due to error {}".format(e.filename, str(e)))

    for dest_path in targets:
        # make a module manifest
        port = dest_path.split(os.path.sep)[-2]
        # todo: add board / variant into manifest files ?
        utils.make_manifest(Path(dest_path), family=FAMILY, port=port, version=version)
    return targets


def get_target_names(path: str) -> tuple:
    "get path to port and board names from a path"
    # https://regexr.com/4sram
    # but with an extra P for Python named groups...
    # regex_1 = r"(?P<port>.*[\\/]+ports[\\/]+\w+)[\\/]+boards[\\/]+\w+"              # port
    re_portboard = r".*[\\/]+ports[\\/]+(?P<port>\w+)[\\/]+boards[\\/]+(?P<board>\w+)"  # port & board
    re_port = r".*[\\/]+ports[\\/]+(?P<port>\w+)[\\/]+"  # port
    # matches= re.search(regex, 'C:\\develop\\MyPython\\micropython\\ports\\esp32\\boards\\TINYPICO\\manifest.py')
    # print( matches.group('port'), matches.group('board'))

    mpy_port = mpy_board = None
    matches = re.search(re_portboard, path)
    if matches:
        # port and board
        mpy_port = matches.group("port") or None
        mpy_board = matches.group("board") or None
    else:
        # just port
        matches = re.search(re_port, path)
        if matches:
            mpy_port = matches.group("port") or None
    return mpy_port, mpy_board


def get_frozen_manifest(
    manifests,
    stub_path: str,
    mpy_path: str,
    lib_path: str,
    version: str,
):
    """
    get and parse the to-be-frozen .py modules for micropython to extract the static type information
    locates the to-be-frozen files through the manifest.py introduced in MicroPython 1.12
    - manifest.py is used for board specific and daily builds
    - manifest_release.py is used for the release builds
    """

    global path_vars  # pylint: disable=global-statement
    global stub_dir  # pylint: disable=global-statement

    stub_path = os.path.abspath(stub_path)

    path_vars["MPY_DIR"] = mpy_path
    path_vars["MPY_LIB_DIR"] = lib_path

    # https://regexr.com/4rh39
    # but with an extra P for Python named groups...
    regex_1 = r"(?P<port>.*[\\/]+ports[\\/]+\w+)[\\/]+boards[\\/]+\w+"  # port
    regex_2 = r"(?P<board>(?P<port>.*[\\/]+ports[\\/]+\w+)[\\/]+boards[\\/]+\w+)"  # port & board

    # todo: variants
    # regex_3 = r"(?P<board>(?P<port>.*[\\/]+ports[\\/]+\w+)[\\/]+variants[\\/]+\w+)"  # port & variant

    # matches= re.search(regex, 'C:\\develop\\MyPython\\micropython\\ports\\esp32\\boards\\TINYPICO\\manifest.py')
    # print( matches.group('port'), matches.group('board'))

    # Include top-level inputs, to generate the manifest
    for manifest in manifests:
        log.info("Manifest: {}".format(manifest))
        path_vars["PORT_DIR"] = ""
        path_vars["BOARD_DIR"] = ""
        matches = re.search(regex_2, manifest)  # BOARD AND PORT
        if matches:
            # port and board
            path_vars["PORT_DIR"] = matches.group("port") or ""
            path_vars["BOARD_DIR"] = matches.group("board") or ""
            if os.path.basename(matches.group("board")) == "manifest":
                path_vars["BOARD_DIR"] = ""
        else:
            # TODO: Variants
            matches = re.search(regex_2, manifest)  # BOARD AND VARIANT
            if matches:
                # port and variant
                path_vars["PORT_DIR"] = matches.group("port") or ""
                path_vars["BOARD_DIR"] = matches.group("board") or ""
                if os.path.basename(matches.group("board")) == "manifest":
                    path_vars["BOARD_DIR"] = ""
            else:
                # just port
                matches = re.search(regex_1, manifest)
                if matches:
                    path_vars["PORT_DIR"] = matches.group("port") or ""

        port_name = os.path.basename(path_vars["PORT_DIR"])
        board_name = os.path.basename(path_vars["BOARD_DIR"])

        if board_name == "":
            board_name = "GENERIC"

        if board_name == "manifest_release":
            board_name = "RELEASE"

        # set global for later use - must be an absolute path.
        stub_dir = os.path.abspath(os.path.join(stub_path, port_name, board_name))

        try:
            include(manifest)
        except FreezeError as er:
            log.error('freeze error executing "{}": {}'.format(manifest, er.args[0]))

        # make a module manifest
        utils.make_manifest(Path(stub_dir), FAMILY, "frozen", version)


if __name__ == "__main__":
    "just gather for the current version"
    logging.basicConfig(format="%(levelname)-8s:%(message)s", level=logging.INFO)
    mpy_path = "./micropython"
    lib_path = "./micropython-lib"
    version = utils.clean_version(git.get_tag(mpy_path) or "x.x")

    if version:
        log.info("found micropython version : {}".format(version))
        # folder/{family}_{version}_frozen
        stub_path = utils.stubfolder("{}-{}-frozen".format(FAMILY, utils.flat_version(version)))
        get_frozen(stub_path, version=version, mpy_path=mpy_path, lib_path=lib_path)
        exit(0)
    else:
        log.warning("Unable to find the micropython repo in folder : {}".format(mpy_path))
        exit(1)



# -*- coding: utf-8 -*-


from contextlib import AbstractContextManager, ExitStack, contextmanager
from pathlib import Path
from typing import List, Optional, Tuple

from boltons import fileutils
from micropy import utils
from micropy.logger import Log, ServiceLog

from .package import Package


class DependencySource(AbstractContextManager):
    """Base class for managing dependency sources.

    Args:
        package (Package): package the source points too.

    """

    _ignore_stubs: List[str] = ["setup.py", "__version__", "test_"]

    def __init__(self, package: Package):
        self.is_local = False
        self._package = package
        self.log: ServiceLog = Log.add_logger(repr(self))

    @property
    def package(self) -> Package:
        return self._package

    @contextmanager
    def handle_cleanup(self):
        with ExitStack() as stack:
            stack.push(self)
            yield
            # no errors, continue on
            stack.pop_all()

    def get_root(self, path: Path) -> Optional[Path]:
        """Determines package root if it has one.

        Args:
            path (Path): Path to check

        Returns:
            bool: True if is package

        """
        init = next(path.rglob("__init__.py"), None)
        if init:
            return init.parent
        return None

    def generate_stubs(self, path: Path) -> List[Tuple[Path, Path]]:
        """Generate Stub Files from a package.

        Args:
            path (Path): Path to package.

        Returns:
            List[Tuple[Path, Path]]: List of tuples containing
                 a path to the original file and stub, respectively.

        """
        py_files = fileutils.iter_find_files(str(path), patterns="*.py", ignored=self._ignore_stubs)
        stubs = [utils.generate_stub(f) for f in py_files]
        return stubs

    def __enter__(self):
        """Method to prepare source."""

    def __exit__(self, *args):
        return super().__exit__(*args)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.package}>"



# -*- coding: utf-8 -*-

"""Project Stubs Module."""

import sys
from pathlib import Path
from typing import Any, List, Sequence, Union

from boltons import setutils
from micropy.project.modules import ProjectModule
from micropy.stubs import StubManager
from micropy.stubs.stubs import DeviceStub


class StubsModule(ProjectModule):
    """Project module for handling Stubs.

    Args:
        stub_manager (StubManager): StubManager instance.
        stubs (List[Type[Stub]], optional): Initial Stubs to use.

    """

    PRIORITY: int = 9

    def __init__(
        self, stub_manager: StubManager, stubs: Sequence[DeviceStub] = None, **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.stub_manager: StubManager = stub_manager
        self._stubs: Sequence[DeviceStub] = stubs or []

    @property
    def context(self):
        """Component stub context."""
        return self.parent.context

    @property
    def config(self) -> dict:
        """Component specific config values.

        Returns:
            dict: Current config.

        """
        return self.parent.config

    @property
    @ProjectModule.hook()
    def stubs(self) -> Union[StubManager, Sequence[DeviceStub]]:
        """Component stubs.

        Returns:
            List[micropy.stubs.Stub]: List of stubs used in project.

        """
        _stubs = self.context.get("stubs", [])
        return self._resolve_subresource(_stubs)

    def get_stub_tree(self, stubs) -> Sequence[Path]:
        """Retrieve and order paths to base stubs and any stubs they depend on.

        Args:
            stubs: List of Stub Items

        Returns:
            Paths to all stubs project depends on.

        """
        stub_tree = setutils.IndexedSet()
        base_stubs = setutils.IndexedSet([s.stubs for s in stubs])
        frozen = [s.frozen for s in stubs]
        fware_mods = [s.firmware.frozen for s in stubs if s.firmware is not None]
        stub_tree.update(*frozen, *fware_mods, *base_stubs)
        return list(stub_tree)

    def _resolve_subresource(
        self, stubs: List[DeviceStub]
    ) -> Union[StubManager, Sequence[DeviceStub]]:
        """Resolves stub resource.

        Args:
            stubs (stubs): Stubs Passed to Manager

        """
        if not hasattr(self, "_parent"):
            return self._stubs
        if not self.parent.exists:
            return self._stubs
        try:
            resource = set(self.stub_manager.resolve_subresource(stubs, self.parent.data_path))
        except OSError as e:
            self.log.error("Failed to Create Stub Links!", exception=e)
            sys.exit(1)
        else:
            self.config.upsert("stubs", {s.name: s.stub_version for s in stubs})
            return resource

    def _load_stub_data(self, stub_data=None, **kwargs):
        """Loads Serialized Stub Data.

        Args:
            stub_data (dict): Dict of Stubs

        """
        for name, location in stub_data.items():
            _path = Path(location).absolute()
            if Path(_path).exists():
                yield self.stub_manager.add(_path)
            yield self.stub_manager.add(name)

    def load(self, **kwargs):
        """Loads stubs from info file.

        Args:
            stub_list (dict): Dict of Stubs

        """
        self.config.upsert("stubs", {s.name: s.stub_version for s in self._stubs})
        stubs = list(self._load_stub_data(stub_data=self.config.get("stubs")))
        stubs.extend(self.stubs)
        stubs = self._resolve_subresource(stubs)
        self.context.upsert("stubs", stubs)
        self.context.upsert("paths", self.get_stub_tree(self.stubs))
        return self.stubs

    def create(self):
        """Create stub project files."""
        self.log.info(f"Stubs: $[{' '.join(str(s) for s in self.stubs)}]")
        return self.load()

    def update(self):
        """Update current project stubs."""
        self.load()
        return self.stubs

    @ProjectModule.hook()
    def add_stub(self, stub, **kwargs):
        """Add stub to project.

        Args:
            stub (Stub): Stub object to add

        Returns:
            [Stubs]: Project Stubs

        """
        self.context.extend("stubs", [stub])
        self.log.info("Loading project...")
        self._resolve_subresource(self.stubs)
        self.log.info("Updating Project Info...")
        self.parent.update()
        self.log.info(f"Project Stubs: $[{' '.join(str(s) for s in self.stubs)}]")
        self.log.success("\nProject Updated!")
        return self.stubs



# -*- coding: utf-8 -*-

import json
import shutil
from copy import deepcopy
from pathlib import Path

import pytest
from micropy import project


@pytest.fixture
def mock_requests(mocker, requests_mock, test_archive):
    mock_source = {
        "name": "Micropy Stubs",
        "location": "https://codeload.github.com/BradenM/micropy-stubs",
        "source": "https://raw.githubusercontent.com/bradenm/micropy-stubs/source.json",
        "path": "legacy.tar.gz/pkg/",
        "packages": [
            {
                "name": "micropython",
                "type": "firmware",
                "sha256sum": "7ff2cce0237268cd52164b77b6c2df6be6249a67ee285edc122960af869b8ed2",
            },
        ],
    }
    requests_mock.get(
        "https://raw.githubusercontent.com/BradenM/micropy-stubs/master/source.json",
        json=mock_source,
    )
    requests_mock.get(
        "https://codeload.github.com/BradenM/micropy-stubs/legacy.tar.gz/pkg/micropython",
        content=test_archive,
    )


@pytest.mark.skip(reason="Tests need some serious cleanup before something like this could work.")
@pytest.mark.usefixtures("mock_requests")
class TestCreateProject:
    mp = None

    expect_mp_data = staticmethod(
        lambda name: {
            "name": "NewProject",
            "stubs": {name: "1.2.0"},
            "packages": {},
            "dev-packages": {"micropy-cli": "*"},
            "config": {"vscode": True, "pylint": True},
        }
    )

    expect_vsc_data = staticmethod(
        lambda name: [
            str(Path(f".micropy/{name}/frozen")),
            str(Path(".micropy/fware_test_stub/frozen")),
            str(Path(f".micropy/{name}/stubs")),
            str(Path(".micropy/NewProject")),
        ]
    )

    def build_project(self, mpy, path):
        proj_path = path / "highlevel_new_project"
        if proj_path.exists():
            shutil.rmtree(proj_path, ignore_errors=True)
        proj = project.Project(proj_path)
        proj_stub = list(mpy.stubs)[0]
        proj.add(project.modules.StubsModule, mpy.stubs, stubs=[proj_stub])
        proj.add(project.modules.PackagesModule, "requirements.txt")
        proj.add(project.modules.DevPackagesModule, "dev-requirements.txt")
        proj.add(project.modules.TemplatesModule, ("vscode", "pylint"))
        return (proj, mpy, proj_stub)

    def check_mp_data(self, path, utils, name="esp32", expect=None):
        expect_data = expect or self.expect_mp_data(name)
        micropy_file = path
        assert micropy_file.exists()
        mp_data = json.loads(micropy_file.read_text())
        assert utils.dict_equal(mp_data, expect_data)

    def check_vscode(self, path, name="esp32", expect=None):
        vsc_path = path / ".vscode" / "settings.json"
        assert vsc_path.exists()
        with vsc_path.open() as f:
            lines = [l.strip() for l in f.readlines() if l]
            valid = [l for l in lines if "//" not in l[:2]]
        vsc_data = json.loads("\n".join(valid))
        expect_data = expect or self.expect_vsc_data(name)
        assert vsc_data["python.analysis.typeshedPaths"] == expect_data

    def test_setup_stubs(self, mock_micropy, get_stub_paths, shared_datadir):
        mpy = mock_micropy
        stub_path = shared_datadir / "esp32_test_stub"
        mpy.stubs.add(stub_path)

    def test_create_project(self, micropy_stubs, tmp_path, utils):
        proj, mpy, proj_stub = self.build_project(micropy_stubs(), tmp_path)
        proj.create()
        self.check_mp_data(proj.info_path, utils, name=proj_stub.path.name)
        self.check_vscode(proj.path, name=proj_stub.path.name)

    def test_add_package(self, mock_pkg, micropy_stubs, tmp_path, utils):
        proj, mpy, proj_stub = self.build_project(micropy_stubs(), tmp_path)
        proj.create()
        proj.add_package("newpackage")
        expect_data = deepcopy(self.expect_mp_data(proj_stub.path.name))
        expect_data["packages"]["newpackage"] = "*"
        self.check_mp_data(proj.info_path, utils, expect=expect_data)

    @pytest.mark.parametrize("local_pkg", ["src/lib/coolpackage", "/tmp/absolute/package"])
    def test_add_local_package(self, tmp_path, local_pkg, micropy_stubs, utils):
        proj, mpy, proj_stub = self.build_project(micropy_stubs(), tmp_path)
        proj.create()
        local_package = Path(local_pkg)
        if not local_package.is_absolute():
            local_package = proj.path / Path(local_pkg)
        local_package.mkdir(parents=True, exist_ok=True)
        (local_package / "__init__.py").touch()
        local_path = utils.str_path(local_pkg)
        proj.add_package(f"-e {local_path}")
        # check micropy.json
        expect_data = deepcopy(self.expect_mp_data(proj_stub.path.name))
        expect_data["packages"][local_package.name] = f"-e {local_path}"
        self.check_mp_data(proj.info_path, utils, expect=expect_data)
        # check vscode settings
        expect_vscode = deepcopy(self.expect_vsc_data(proj_stub.path.name))
        expect_vscode.append(local_path)
        self.check_vscode(proj.path, expect=expect_vscode)
        shutil.rmtree(proj.path)



# make_stub_files: Thu 20 Jun 2019 at 23:08:04

from typing import Any, Dict, Optional, Sequence, Tuple, Union

Node = Any
def roundup(val: Any, align: Any) -> Any: ...
    #   0: return val+align-1&~align-1
    # ? 0: return val+align-number&align-number
class FileSection:
    def __init__(self, f: Any, content_len: smallint, aligned_len: smallint) -> None: ...
    def read(self, sz: Any=65536) -> Union[Any, bytes]: ...
        #   0: return b''
        #   0: return bytes
        #   1: return data
        # ? 1: return data
    def readinto(self, buf: Any) -> Union[Any, number]: ...
        #   0: return 0
        #   0: return number
        #   1: return sz
        # ? 1: return sz
    def skip(self) -> None: ...
class TarInfo:
    def __str__(self) -> str: ...
class TarFile:
    def __init__(self, name: str=None, fileobj: Any=None) -> None: ...
    def next(self) -> Optional[Any]: ...
        #   0: return None
        #   0: return None
        #   1: return None
        #   1: return None
        #   2: return d
        # ? 2: return d
    def __iter__(self) -> Any: ...
        #   0: return self
        # ? 0: return self
    def __next__(self) -> Any: ...
        #   0: return v
        # ? 0: return v
    def extractfile(self, tarinfo: Any) -> Any: ...
        #   0: return tarinfo.subf
        # ? 0: return tarinfo.subf



# -*- coding: utf-8 -*-

import json

import pytest
from micropy import config


class TestConfig:

    default = {"one": 1, "two": 2, "sub": {"items": ["foo", "bar"], "bool": True}}

    @pytest.fixture
    def test_config(self, tmp_path):
        cfg_file = tmp_path / "conf.json"
        conf = config.Config(cfg_file, default=self.default)
        return conf

    def get_file_data(self, conf):
        path = conf.source.file_path
        return json.loads(path.read_text())

    def test_default(self, tmp_path):
        cfg_file = tmp_path / "conf.json"
        conf = config.Config(cfg_file, default=self.default)
        # should not create source file until first change
        assert not cfg_file.exists()
        # make change
        conf.set("one", 2)
        assert cfg_file.exists()
        assert self.get_file_data(conf) == conf.raw()

    def test_load_from_file(self, tmp_path, utils):
        cfg_file = tmp_path / "conf.json"
        cfg_file.write_text(json.dumps(self.default))
        conf = config.Config(cfg_file, default={})
        assert conf.config == self.default
        # default should be overriden
        conf = config.Config(cfg_file, default={"one": 1})
        assert utils.dict_equal(conf.raw(), self.default)

    def test_override(self, test_config, tmp_path):
        conf = test_config
        new_cfg = tmp_path / "newcfg.json"
        conf.source = new_cfg
        assert isinstance(conf.source, config.JSONConfigSource)
        diff_cfg = tmp_path / "diffcfg.json"
        conf.source.file_path = diff_cfg
        conf.sync()
        assert self.get_file_data(conf) == conf.config

    def test_get(self, test_config):
        conf = test_config
        assert conf.get("one") == 1
        assert conf.get("sub/bool")
        assert conf.get("sub/items/0") == "foo"
        assert conf.get("sub/items") == ["foo", "bar"]

    def test_set(self, test_config):
        conf = test_config
        conf.set("one", 1)
        conf.set("one/sub/items.0", "foobar")
        data = json.loads(conf.source.file_path.read_text())
        assert data == conf.config

    def test_update_from_file(self, test_config):
        conf = test_config
        cfg_file = conf.source.file_path
        new = self.default.copy()
        new["one"] = 45
        new["section"] = {"value": "foo"}
        cfg_file.write_text(json.dumps(new))
        conf = config.Config(cfg_file, default=self.default)
        assert conf.get("one") == 45
        assert conf.get("section/value") == "foo"
        assert conf.config == new

    def test_extend(self, test_config):
        conf = test_config
        conf.extend("sub/items", ["foobar", "barfoo"])
        file_data = json.loads(conf.source.file_path.read_text())
        print(file_data)
        assert file_data["sub"]["items"] == ["foo", "bar", "foobar", "barfoo"]
        assert conf.get("sub/items") == ["foo", "bar", "foobar", "barfoo"]

    def test_upsert(self, test_config):
        conf = test_config
        conf.upsert("sub/items", ["barfoo", "foobar", "bar", "foo"])
        file_data = self.get_file_data(conf)
        assert file_data["sub"]["items"] == ["barfoo", "foobar", "bar", "foo"]
        assert conf.get("sub/items") == ["barfoo", "foobar", "bar", "foo"]

    def test_dict(self):
        conf = config.Config(source_format=config.DictConfigSource, default=self.default)
        assert conf.get("one") == 1
        conf.set("sub/bool", False)
        assert not conf.get("sub/bool")



# -*- coding: utf-8 -*-


from pathlib import Path
from typing import List, Optional, Tuple

import requirements
from packaging.utils import canonicalize_name


class Package:
    def __init__(
        self,
        name: str,
        specs: List[Tuple[str, str]],
        path: Optional[Path] = None,
        uri: Optional[str] = None,
        vcs: Optional[str] = None,
        revision: Optional[str] = None,
        line: Optional[str] = None,
        **kwargs,
    ):
        """Generic Python Dependency.

        Args:
            name (str): Name of package
            specs (List[Tuple[str, str]]): Package constraints.
            path: path to package

        """
        self._name = name
        self._specs = specs
        self._path = path
        self._uri = uri
        self._vcs = vcs
        self._revision = revision
        self._line = line
        self.editable = self._path is not None

    @property
    def name(self) -> str:
        return canonicalize_name(self._name)

    @property
    def path(self) -> Optional[Path]:
        if not self._path:
            return None
        return Path(self._path)

    @property
    def full_name(self) -> str:
        if self._line and self._vcs:
            return self._line
        if self._path:
            return self.pretty_specs
        if not self._specs:
            return self.name
        return f"{self.name}{self.pretty_specs}"

    @property
    def uri(self) -> Optional[str]:
        if self._vcs and self._vcs in self._uri[:4]:
            # handle 'git+https' schemas
            return self._uri[4:]
        return self._uri

    @property
    def vcs(self) -> Optional[str]:
        return self._vcs

    @property
    def revision(self) -> Optional[str]:
        return self._revision

    @property
    def line(self) -> Optional[str]:
        return self._line

    @property
    def specs(self) -> List[Tuple[str, str]]:
        return self._specs

    @property
    def pretty_specs(self) -> str:
        if self.line and self.vcs:
            return self.line
        if self._path:
            return f"-e {self._path}"
        if not self.specs:
            return "*"
        _specs = ["".join(i for i in s) for s in self.specs]
        return "".join(_specs)

    @classmethod
    def from_text(cls, name: str, specs: str) -> "Package":
        """Create package from text.

        Args:
            name: name of package
            specs: package constraints

        Returns:
            Package instance

        """
        if "http" in specs:
            req = next(requirements.parse(specs))
            return cls(**req.__dict__)
        if "-e" in specs:
            req = next(requirements.parse(specs))
            return cls(name, req.specs, path=req.path)
        req_name = name
        if specs != "*":
            req_name = f"{name}{specs}"
        req = next(requirements.parse(req_name))
        return cls(req.name, req.specs)

    def __str__(self) -> str:
        return self.full_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Pre/Post Processing for createstubs.py"""
import argparse
import itertools
import sys
from pathlib import Path
import subprocess

try:
    import python_minifier
except ImportError:
    python_minifier = None

ROOT = Path(__file__).parent
source_script = ROOT / "board" / "createstubs.py"
destination_script = ROOT / "minified" / "createstubs.py"


def edit_lines(content, edits, show_diff=False):
    """Edit string by list of edits

    Args:
        content (str): content to edit
        edits ([(str, str)]): List of edits to make.
            The first string in the tuple represents
            the type of edit to make, can be either:
                comment - comment text out (removed on minify)
                rprint - replace text with print
                rpass - replace text with pass
            The second string is the matching text to replace
        show_diff (bool, optional): Prints diff of each edit.
            Defaults to False.

    Returns:
        str: edited string
    """

    def comment(l, x):
        return l.replace(x, f"# {x}")

    def rprint(l, x):  # lgtm [py/unused-local-variable] pylint: disable= unused-variable
        split = l.split("(")
        if len(split) > 1:
            return l.replace(split[0].strip(), "print")
        return l.replace(x, f"print")

    def rpass(l, x):  # lgtm [py/unused-local-variable] pylint: disable= unused-variable
        return l.replace(x, f"pass")

    def get_whitespace_context(content, index):
        """Get whitespace count of lines surrounding index"""

        def count_ws(line):
            return sum(1 for _ in itertools.takewhile(str.isspace, line))

        lines = content[index - 1 : index + 2]
        context = (count_ws(l) for l in lines)
        return context

    def handle_multiline(content, index):
        """Handles edits that require multiline comments

        Example:
            self._log.debug("info: {} {}".format(
                1,
                2
            ))
        Here, only commenting out the first self._log line will raise
        an error. So this function returns all lines that need to
        be commented out instead.

        It also checks for situations such as this:
            if condition:
                self._log.debug('message')

        Here, since the only functionality of the conditional is the call log,
        both lines would be returned to comment out.

        """
        line = content[index]
        open_cnt = line.count("(")
        close_cnt = line.count(")")
        ahead_index = 1
        look_ahead = 0
        while not open_cnt == close_cnt:
            look_ahead = l_index + ahead_index
            ahead_index += 1
            next_l = content[look_ahead]
            open_cnt += next_l.count("(")
            close_cnt += next_l.count(")")
        if ahead_index > 1:
            return range(index, look_ahead + 1)
        prev = content[index - 1]
        _, line_ws, post_ws = get_whitespace_context(content, index)
        prev_words = prev.strip().strip(":").split()
        check = any(
            t
            in (
                "if",
                "else",
            )
            for t in prev_words
        )
        if check and line_ws != post_ws:
            return range(index - 1, index + 1)

    def handle_try_except(content, index):
        """Checks if line at index is in try/except block

        Handles situations like this:
            try:
                something()
            except:
                self._log.debug('some message')

        Simply removing the self._log call would create a syntax error,
        which is what this function checks for.

        """
        prev = content[index - 1]
        _, line_ws, post_ws = get_whitespace_context(content, index)
        if "except" in prev and line_ws != post_ws:
            return True

    lines = []
    multilines = set()
    content = content.splitlines(keepends=True)
    for line in content:
        _line = line
        for edit, text in edits:
            if text in line:
                if edit == "comment":
                    l_index = content.index(line)
                    # Check if edit spans multiple lines
                    mline = handle_multiline(content, l_index)
                    if mline:
                        multilines.update(mline)
                        break
                    # Check if line is only statement in try/except
                    if handle_try_except(content, l_index):
                        edit = "rpass"
                        text = line.strip()
                func = eval(edit)  # pylint: disable= eval-used
                line = func(line, text)
                if line != _line:
                    if show_diff:
                        print(f"\n- {_line.strip()}")
                        print(f"+ {line.strip()}")
                    break
        lines.append(line)
    for line_num in multilines:
        # Go back and comment out multilines
        line = lines[line_num]
        lines[line_num] = comment(line, line.strip())
    stripped = "".join(lines)
    return stripped


def minify_script(source_script: Path, keep_report=True, show_diff=False):
    """minifies createstubs.py

    Args:
        keep_report (bool, optional): Keeps single report line in createstubs
            Defaults to True.
        show_diff (bool, optional): Print diff from edits. Defaults to False.

    Returns:
        str: minified source text
    """

    edits = [
        ("comment", "print"),
        ("comment", "import logging"),
        # first full
        ("comment", "self._log ="),
        ("comment", "self._log("),
        ("comment", "self._log.debug"),
        ("comment", "self._log.warning"),
        ("comment", "self._log.info"),
        ("comment", "self._log.error"),
        # then short versions
        ("comment", "_log ="),
        ("comment", "_log.debug"),
        ("comment", "_log.warning"),
        ("comment", "_log.info"),
        ("comment", "_log.error"),
    ]
    if keep_report:
        report = (
            "rprint",
            ('self._log.info("Stub module: {:<20} to file:' ' {:<55} mem:{:>5}".' "format(module_name, file_name, m1))"),
        )
        clean = (
            "rprint",
            'self._log.info("Clean/remove files in folder: {}".format(path))',
        )
        edits.insert(0, report)
        edits.insert(1, clean)

    with source_script.open("r") as f:
        if not python_minifier:
            raise Exception("python_minifier not available")

        content = f.read()

        content = edit_lines(content, edits, show_diff=show_diff)

        source = python_minifier.minify(
            content,
            filename=source_script.name,
            combine_imports=True,
            remove_literal_statements=True,  # no Docstrings
            remove_annotations=True,  # not used runtime anyways
            hoist_literals=True,  # remove redundant strings
            rename_locals=True,  # short names save memory
            preserve_locals=["stubber"],  # names to keep
            rename_globals=True,  # short names save memory
            # keep these globals to allow testing/mocking to work agains the minified version
            preserve_globals=[
                "main",
                "Stubber",
                "read_path",
                "get_root",
                "_info",
                "os",
                "sys",
                "__version__",
            ],
            # remove_pass=True,  # no dead code
            # convert_posargs_to_args=True, # Does not save any space
        )
    print(f"Original length : {len(content)}")
    print(f"Minified length : {len(source)}")
    print(f"Reduced by      : {len(content)-len(source)} ")
    return source


def cli_minify(**kwargs):
    """minify cli handler"""
    source_path: Path = kwargs.pop("source")
    print(f"\nMinifying {source_path}...")
    out: Path = kwargs.pop("output")
    if not python_minifier:
        print("python_minifier is required to minify createstubs.py\n")
        print("Please install via:\n  pip install python_minifier")
        sys.exit(1)
    with out.open("w+") as f:
        report = kwargs.pop("no_report")
        diff = kwargs.pop("diff")

        source = minify_script(source_script=source_path, keep_report=report, show_diff=diff)
        f.write(source)

    print("Minified file written to :", out)
    result = subprocess.run(["mpy-cross", "-O2", str(out)])
    if result.returncode == 0:
        print("mpy-cross compiled to    :", out.with_suffix(".mpy"))

    print("\nDone!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pre/Post Processing for createstubs.py")

    parser.set_defaults(func=None)
    parser.add_argument(
        "-s",
        "--source",
        help="Specify source file to read from. Defaults to {}".format(source_script),
        type=Path,
        default=source_script,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Specify file to output to. Defaults to {}".format(destination_script),
        type=Path,
        default=destination_script,
    )
    subparsers = parser.add_subparsers(help="Command to execute")

    minify_parser = subparsers.add_parser("minify", help=("Create minified version of createstubs.py"))
    minify_parser.add_argument("-d", "--diff", help="Print diff report from minify", action="store_true")
    minify_parser.add_argument(
        "-n",
        "--no-report",
        help=("Disables all output from createstubs.py." " Use if your having memory related issues."),
        action="store_false",
    )
    minify_parser.set_defaults(func=cli_minify)

    args = parser.parse_args()
    if not args.func:
        parser.print_help()
    else:
        args.func(**vars(args))



# -*- coding: utf-8 -*-

"""Module for handling jinja2 and MicroPy Templates."""

import json
import os
from itertools import chain
from pathlib import Path
from typing import Iterator, List

from jinja2 import Environment, FileSystemLoader
from micropy.logger import Log

from .checks import TEMPLATE_CHECKS


class Template:
    """Base Template Builder Class.

    Args:
        template (jinja2.Template): Jinja2 Template Instance

    Raises:
        NotImplementedError: Method must be overriden by subclass

    """

    FILENAME = None
    CHECKS = []

    def __init__(self, template, **kwargs):
        self.template = template
        self.stubs = kwargs.get("stubs", [])
        self.paths = kwargs.get("paths", [])
        self.datadir = kwargs.get("datadir", None)
        self.local_paths = kwargs.get("local_paths", [])

    @property
    def context(self):
        """Context for template."""
        raise NotImplementedError

    def iter_clean(self, data=None):
        """Yields cleaned data.

        Args:
            data (str, optional): Alternative data to clean.
                Defaults to None. If none, uses template render.

        """
        render = data or self.template.render(self.context)
        for line in render.splitlines(True):
            _line = line.strip()
            if not _line.startswith("//"):
                yield line

    def run_checks(self):
        """Runs all template checks.

        Returns:
            bool: True if all checks passed

        """
        if not self.CHECKS:
            return True
        results = [not ck() for ck in self.CHECKS]
        return any(results)

    def update(self, root):
        """Update Template File.

        Args:
            root (str): Path to project root

        Raises:
            NotImplementedError: Raised if Subclass has not Implemented Update

        Returns:
            func: Template Update Func

        """
        update_func = getattr(self, "update_method", None)
        update_kwargs = getattr(self, "update_kwargs", {})
        if not update_func:
            return None
        path = root / self.FILENAME
        return update_func(path, **update_kwargs)

    def update_as_json(self, path):
        """Update template file as JSON.

        Args:
            path (str): File path to update

        """
        render = json.loads("".join(self.iter_clean()))
        data = json.loads("".join(self.iter_clean(path.read_text())))
        data.update(render)
        with path.open("w+") as f:
            json.dump(data, f, indent=4)

    def update_as_text(self, path, by_contains=None):
        """Update template file as text.

        Args:
            path (str): file path to update.
            by_contains ([str], optional): Update lines that contain a string.
             Defaults to None.

        """
        r_lines = list(self.iter_clean())
        upd_lines = []
        if by_contains:
            upd_lines = [r_lines.index(l) for l in r_lines if any(i in l for i in by_contains)]
        with path.open("r+") as f:
            c_lines = self.iter_clean(f.read())
            f.seek(0)
            for it, line in enumerate(c_lines):
                _line = line
                if it in upd_lines:
                    _line = r_lines[it]
                f.write(_line)

    def render_stream(self):
        """Returns template stream from context."""
        stream = self.template.stream(self.context)
        return stream

    def iter_relative_paths(self, paths: List[Path], strict: bool = False) -> Iterator[Path]:
        """Iterate over a list of paths relative to project root.

        Args:
            paths: List of paths to make relative.
            strict: Raises ValueError if True and path cannot be made relative.
                Defaults to False.

        Raises:
            ValueError: Path could not be made relative and `strict` is True.

        Yields:
            Path relative to project root.

        """
        for p in paths:
            path = p
            if not p.is_absolute():
                path = Path(os.path.relpath(path)).resolve()
            try:
                yield path.relative_to(self.datadir.parent)
            except ValueError:
                yield path

    def __str__(self):
        cls_name = self.__class__.__name__
        return f"{cls_name}[{self.template.name}]::[{self.context}]"


class GenericTemplate(Template):
    """Generic Template for files without context."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.FILENAME = self.template.name

    @property
    def context(self):
        """Empty Context."""
        return {}


class CodeTemplate(Template):
    """Template for VSCode settings."""

    FILENAME = ".vscode/settings.json"
    CHECKS = [TEMPLATE_CHECKS["ms-python"]]

    def __init__(self, *args, **kwargs):
        self.update_method = self.update_as_json
        super().__init__(*args, **kwargs)

    @property
    def context(self):
        """VScode Config Context."""
        paths = self.paths
        if self.datadir:
            paths = list(self.iter_relative_paths(self.paths, strict=True))
        if self.local_paths:
            paths.extend(self.iter_relative_paths(self.local_paths))
        stub_paths = json.dumps([str(s) for s in paths])
        ctx = {
            "stubs": self.stubs or [],
            "paths": stub_paths,
        }
        return ctx


class PylintTemplate(Template):
    """Template for Pylint settings."""

    FILENAME = ".pylintrc"

    def __init__(self, *args, **kwargs):
        self.update_method = self.update_as_text
        self.update_kwargs = {"by_contains": ["sys.path[1:1]"]}
        super().__init__(*args, **kwargs)

    @property
    def context(self):
        """Pylint Config Context."""
        paths = self.paths
        if self.datadir:
            paths = list(self.iter_relative_paths(self.paths, strict=True))
        if self.local_paths:
            paths.extend(self.iter_relative_paths(self.local_paths))
        ctx = {"stubs": self.stubs or [], "paths": paths or []}
        return ctx


class TemplateProvider:
    """Template Factory."""

    _template_files = {
        "vscode": CodeTemplate,
        "pylint": PylintTemplate,
        "vsextensions": ".vscode/extensions.json",
        "pymakr": "pymakr.conf",
        "main": "src/main.py",
        "boot": "src/boot.py",
        "gitignore": ".gitignore",
    }
    ENVIRONMENT = None
    TEMPLATE_DIR = Path(__file__).parent / "template"
    TEMPLATES = {
        "vscode": (
            ["vscode", "vsextensions"],
            ("VSCode Settings for " "Autocompletion/Intellisense"),
        ),
        "pymakr": (["pymakr"], "Pymakr Configuration"),
        "pylint": (["pylint"], "Pylint MicroPython Settings"),
        "gitignore": (["gitignore"], "Git Ignore Template"),
        "bootstrap": (["main", "boot"], "main.py & boot.py files"),
    }

    def __init__(self, templates, log=None, **kwargs):
        """Template Factory.

        Args:
            templates ([str]): List of Templates to use
            log (callable, optional): Log instance to use.
                Defaults to None. If none, creates a new one.
            run_checks (bool, optional): Whether to run template checks.
                Defaults to True.

        """
        self.run_checks = kwargs.get("run_checks", True)
        self.template_names = set(
            chain.from_iterable([self.TEMPLATES.get(t)[0] for t in templates])
        )
        self.files = {k: v for k, v in self._template_files.items() if k in self.template_names}
        self.log = log or Log.add_logger("Templater")
        if self.__class__.ENVIRONMENT is None:
            loader = FileSystemLoader(str(self.TEMPLATE_DIR))
            self.__class__.ENVIRONMENT = Environment(loader=loader)
            self.log.debug("Created Jinja2 Environment")
            self.log.debug(f"Detected Templates: {self.ENVIRONMENT.list_templates()}")

    def get(self, name, *args, **kwargs):
        """Retrieve appropriate Template instance by name.

        Args:
            name (str): Name of template

        Returns:
            Template: Template instance

        """
        temp_def = self.files.get(name)
        file_attr = getattr(temp_def, "FILENAME", None)
        filename = temp_def if file_attr is None else file_attr
        temp_cls = GenericTemplate if file_attr is None else temp_def
        file_temp = self.ENVIRONMENT.get_template(filename)
        self.log.debug(f"Retrieving {name} as {temp_cls} from {file_temp.name}")
        template = temp_cls(file_temp, *args, **kwargs)
        return template

    def render_to(self, name, parent_dir, *args, **kwargs):
        """Renders Template to a file under parent directory.

        Args:
            name (str): Name of template
            parent_dir (str): Path to root dir

        """
        template = self.get(name, **kwargs)
        self.log.debug(f"Loaded: {str(template)}")
        if self.run_checks:
            self.log.debug(f"Verifying {template} requirements...")
            template.run_checks()
        out_dir = parent_dir / template.FILENAME
        if not os.path.isfile(out_dir):
            self.log.debug(f"Create: {out_dir}")
            parent_dir.mkdir(exist_ok=True)
            out_dir.parent.mkdir(exist_ok=True, parents=True)
            self.log.debug(f"Rendered: {name} to {str(out_dir)}")
            self.log.info(f"$[{name.capitalize()}] File Generated!")
            stream = template.render_stream()
            return stream.dump(str(out_dir))
        else:
            self.log.debug(f"Update: {out_dir}")
            template.update(parent_dir)
            self.log.info(f"$[{name.capitalize()}] File Updated!")

    def update(self, name, root_dir, **kwargs):
        """Update existing Template.

        Args:
            name (str): Template name
            root_dir (str): Path to project root

        Returns:
            Template: Updated Template Instance

        """
        template = self.get(name, **kwargs)
        self.log.debug(f"Loaded: {str(template)}")
        try:
            template.update(root_dir)
        except FileNotFoundError:
            self.log.debug("Template does not exist!")
            return self.render_to(name, root_dir, **kwargs)
        self.log.debug(f"Updated: {str(template)}")
        return template

    @property
    def templates(self):
        """returns all template names."""
        return self.files.keys()



# (c) 2019 Paul Sokolovsky, MIT license
from token import *

from ucollections import namedtuple

NL = 55
ENCODING = 56
tok_name[NL] = "NL"
tok_name[ENCODING] = "ENCODING"


class TokenInfo(namedtuple("TokenInfo", ("type", "string", "start", "end", "line"))):
    def __str__(self):
        return "TokenInfo(type=%d (%s), string=%r, line=%r)" % (
            self.type,
            tok_name[self.type],
            self.string,
            self.line,
        )


def get_indent(l):
    for i in range(len(l)):
        if l[i] != " ":
            return i, l[i:]


def tokenize(readline):

    indent = 0

    yield TokenInfo(ENCODING, "utf-8", 0, 0, "")

    while True:
        l = readline()
        org_l = l
        if not l:
            break
        i, l = get_indent(l)

        if l == "\n":
            yield TokenInfo(NL, l, 0, 0, org_l)
            continue

        if i > indent:
            yield TokenInfo(INDENT, " " * (i - indent), 0, 0, org_l)
        elif i < indent:
            yield TokenInfo(DEDENT, "", 0, 0, org_l)
        indent = i

        while l:
            if l[0].isdigit():
                t = ""
                while l and (l[0].isdigit() or l[0] == "."):
                    t += l[0]
                    l = l[1:]
                yield TokenInfo(NUMBER, t, 0, 0, org_l)
            elif l[0].isalpha():
                name = ""
                while l and (l[0].isalpha() or l[0].isdigit()):
                    name += l[0]
                    l = l[1:]
                yield TokenInfo(NAME, name, 0, 0, org_l)
            elif l[0] == "\n":
                yield TokenInfo(NEWLINE, "\n", 0, 0, org_l)
                break
            elif l[0].isspace():
                l = l[1:]
            else:
                yield TokenInfo(OP, l[0], 0, 0, org_l)
                l = l[1:]

    yield TokenInfo(ENDMARKER, "", 0, 0, "")



# -*- coding: utf-8 -*-

"""Project Templates Module."""


from micropy.project.modules import ProjectModule
from micropy.project.template import TemplateProvider


class TemplatesModule(ProjectModule):
    """Project Templates Module.

    Generates and manages project files using the Projects
    context.

    Args:
        templates (List[str]): List of templates to use.
        run_checks (bool, optional): Whether to execute checks or not.
            Defaults to True.

    """

    PRIORITY: int = 1
    TEMPLATES = TemplateProvider.TEMPLATES
    _dynamic = ["vscode", "pylint"]

    def __init__(self, templates=None, run_checks=True, **kwargs):
        self._templates = templates or []
        super().__init__(**kwargs)
        self.run_checks = run_checks

    @property
    def config(self):
        """Template config.

        Returns:
            dict: Current configuration

        """
        return self.parent.config

    def get_provider(self, templates):
        return TemplateProvider(templates, run_checks=self.run_checks, log=self.log)

    def load(self, **kwargs):
        """Loads project templates."""
        self.provider = self.get_provider(self.config.get("config"))
        templates = [k for k, v in self.config.get("config").items() if v]
        self.log.debug(f"Loading Templates: {templates}")
        self.provider = TemplateProvider(templates, **kwargs)
        self.update()

    def create(self):
        """Generates project files.

        Returns:
            dict: Project context

        """
        self.log.title("Rendering Templates")
        self.log.info("Populating Stub Info...")
        for key in self._templates:
            if key in self._dynamic:
                self.config.add("config" + "/" + key, True)
        self.provider = self.get_provider(self._templates)
        for t in self.provider.templates:
            self.provider.render_to(t, self.parent.path, **self.parent.context.raw())
        self.log.success("Stubs Injected!")
        return self._templates

    def update(self):
        """Updates project files.

        Returns:
            dict: Project context

        """
        self.provider = self.get_provider(self.config.get("config"))
        self.log.debug(f"updating templates with context: {self.parent.context.raw()}")
        for tmp in self.provider.templates:
            self.provider.update(tmp, self.parent.path, **self.parent.context.raw())
        return self.parent.context



# -*- coding: utf-8 -*-


import io

import pytest
import requests
from jsonschema import ValidationError
from micropy import utils
from requests.exceptions import ConnectionError, HTTPError, InvalidURL


@pytest.fixture
def schema(datadir):
    file = datadir / "schema.json"
    pass_file = datadir / "pass.json"
    fail_file = datadir / "fail.json"
    return (file, pass_file, fail_file)


def test_validate(schema):
    """Test for successful validation"""
    schema, pass_file, _ = schema
    val = utils.Validator(schema_path=schema)
    val.validate(pass_file)


def test_fail_validate(schema):
    """Test for invalid file"""
    schema, _, fail_file = schema
    val = utils.Validator(schema_path=schema)
    with pytest.raises(ValidationError):
        val.validate(fail_file)


def test_is_url(test_urls):
    """should respond true/false for url"""
    u = test_urls
    assert utils.is_url(u["valid"])
    assert utils.is_url(u["valid_https"])
    assert not utils.is_url(u["invalid"])
    assert not utils.is_url(u["invalid_file"])


def test_ensure_valid_url(mocker, test_urls):
    """should ensure url is valid"""
    u = test_urls
    with pytest.raises(InvalidURL):
        utils.ensure_valid_url(test_urls["invalid"])
    with pytest.raises(ConnectionError):
        mocker.patch.object(utils, "is_url", return_value=True)
        mock_head = mocker.patch.object(requests, "head")
        mock_head.side_effect = [ConnectionError]
        utils.ensure_valid_url(u["valid"])
    mocker.stopall()
    with pytest.raises(HTTPError):
        utils.ensure_valid_url(u["bad_resp"])
    result = utils.ensure_valid_url(u["valid"])
    assert result == u["valid"]


def test_ensure_existing_dir(tmp_path):
    """should ensure dir exists"""
    not_exist = tmp_path / "i_dont_exist"
    file = tmp_path / "file.txt"
    file.touch()
    with pytest.raises(NotADirectoryError):
        utils.ensure_existing_dir(not_exist)
    with pytest.raises(NotADirectoryError):
        utils.ensure_existing_dir(file)
    result = utils.ensure_existing_dir(str(tmp_path))
    assert result == tmp_path
    assert result.exists()
    assert result.is_dir()


def test_is_downloadable(mocker, test_urls):
    """should check if url can be downloaded from"""
    u = test_urls
    uheaders = u["headers"]
    mock_head = mocker.patch.object(requests, "head")
    head_mock_val = mocker.PropertyMock(
        side_effect=[uheaders["not_download"], uheaders["can_download"]]
    )
    type(mock_head.return_value).headers = head_mock_val
    assert not utils.is_downloadable(u["valid"])
    assert not utils.is_downloadable("not-a-real-url")
    assert utils.is_downloadable(u["valid"])


def test_get_url_filename(test_urls):
    """should return filename"""
    filename = "archive_test_stub.tar.gz"
    result = utils.get_url_filename(test_urls["download"])
    assert result == filename


def test_is_existing_dir(tmp_path):
    bad_path = tmp_path / "not-real-path"
    is_file = tmp_path / "file.txt"
    is_file.touch()
    assert not utils.is_existing_dir(bad_path)
    assert not utils.is_existing_dir(is_file)
    assert utils.is_existing_dir(tmp_path)


def test_search_xml(mocker, shared_datadir, test_urls):
    u = test_urls
    test_xml = shared_datadir / "test_source.xml"
    mock_get = mocker.patch.object(requests, "get")
    with test_xml.open("rb") as f:
        type(mock_get.return_value).content = f.read()
    results = utils.search_xml(u["valid"], "Key", ignore_cache=True)
    assert sorted(results) == sorted(
        ["packages/esp32-micropython-1.10.0.tar.gz", "packages/esp32-micropython-1.11.0.tar.gz"]
    )


def test_generate_stub(shared_datadir, tmp_path, mocker):
    mock_stubber = mocker.patch.object(utils.stub, "import_stubber").return_value
    mock_stubber.__file__ = tmp_path
    expect_path = tmp_path / "foo.py"
    expect_path.touch()
    result = utils.generate_stub(expect_path)
    mock_stubber.generate_pyi_from_file.assert_called_once()
    assert result == (expect_path, expect_path.with_suffix(".pyi"))
    # Test print monkeypatch
    print_mock = mocker.Mock(return_value=None)
    utils.generate_stub(expect_path, log_func=print_mock)


def test_get_package_meta(mocker, requests_mock):
    """should get package meta"""
    mock_data = {
        "releases": {
            "0.0.0": [{"url": "early-version.tar.gz"}],
            "0.1.0": [
                {
                    "url": "do-not-return-me",
                },
                {"url": "return-me.tar.gz"},
            ],
        }
    }
    requests_mock.get("https://pypi.org/pypi/foobar/json", json=mock_data)
    result = utils.get_package_meta("foobar", "https://pypi.org/pypi/foobar/json")
    assert result == {"url": "return-me.tar.gz"}
    result = utils.get_package_meta("foobar==0.0.0", "https://pypi.org/pypi/foobar/json")
    assert result == {"url": "early-version.tar.gz"}


def test_extract_tarbytes(mocker):
    """should extract tar file from memory"""
    test_bytes = bytearray("foobar", "utf-8")
    mock_io = mocker.patch.object(utils.helpers.io, "BytesIO")
    mock_io.return_value = io.BytesIO(test_bytes)
    mock_tarfile = mocker.patch.object(utils.helpers, "tarfile")
    mock_tar = mock_tarfile.open.return_value.__enter__.return_value
    utils.extract_tarbytes(test_bytes, "foobar")
    mock_tarfile.open.assert_called_once_with(fileobj=io.BytesIO(test_bytes), mode="r:gz")
    mock_tar.extractall.assert_called_once_with("foobar")


def test_iter_requirements(mocker, tmp_path):
    """should iter requirements"""
    tmp_file = tmp_path / "tmp_reqs.txt"
    tmp_file.touch()
    tmp_file.write_text("micropy-cli==1.0.0")
    result = next(utils.iter_requirements(tmp_file))
    assert result.name == "micropy-cli"
    assert result.specs == [("==", "1.0.0")]


def test_create_dir_link(mocker, tmp_path):
    """Should create a symlink or directory junction if needed"""
    targ_path = tmp_path / "target_dir"
    targ_path.mkdir()
    link_path = tmp_path / "link_path"
    mock_sys = mocker.patch.object(utils.helpers, "sys")
    mock_platform = type(mock_sys).platform = mocker.PropertyMock()
    mock_subproc = mocker.patch.object(utils.helpers, "subproc")
    mock_path = mocker.patch.object(utils.helpers, "Path").return_value
    mock_path.symlink_to.side_effect = [mocker.ANY, OSError, OSError, OSError]

    mock_platform.return_value = "linux"
    # Test POSIX (should not raise exception)
    utils.create_dir_link(link_path, targ_path)
    mock_path.symlink_to.assert_called_once()
    assert mock_subproc.call_count == 0
    # Test POSIX failed for unknown reason
    with pytest.raises(OSError):
        utils.create_dir_link(link_path, targ_path)
    # Test Windows (should try to make symlink, fallback on DJ)
    mock_platform.return_value = "win32"
    mock_subproc.call.return_value = 0
    utils.create_dir_link(link_path, targ_path)
    assert mock_subproc.call.call_count == 1
    # Test Windows fails for some reason
    mock_subproc.call.return_value = 1
    with pytest.raises(OSError):
        utils.create_dir_link(link_path, targ_path)


def test_is_dir_link(mocker, tmp_path):
    """Should test if a path is a symlink or directory junction"""
    link_path = tmp_path / "link"
    targ_path = tmp_path / "target"
    mock_sys = mocker.patch.object(utils.helpers, "sys")
    mock_platform = type(mock_sys).platform = mocker.PropertyMock()
    mock_path = mocker.patch.object(utils.helpers, "Path").return_value
    mock_path.is_symlink.side_effect = [True, False, False, False]
    # Test Symlink (POSIX)
    mock_platform.return_value = "linux"
    assert utils.is_dir_link(link_path)
    assert not utils.is_dir_link(link_path)
    # Test Directory Junction (Windows)
    mock_platform.return_value = "win32"
    # From what I can tell, while Path.is_symlink always returns false for DJs.
    # However, on a DJ, Path.absolute will return the absolute path to the DJ,
    # while Path.resolve will return the absolute path to the source directory.
    # With this in mind, this check SHOULD work.
    mock_path.resolve.return_value = targ_path
    mock_path.absolute.return_value = link_path
    assert utils.is_dir_link(link_path)
    mock_path.absolute.return_value = targ_path
    assert not utils.is_dir_link(link_path)


@pytest.mark.parametrize(
    "versions,expect",
    [
        (["1.0.0rc.1"], False),
        (["1.0.0rc.1", "1.0.0"], "1.0.0"),
        (["1.0.0rc.1", "1.0.0", "2.0.0rc.1", "2.0.0"], "2.0.0"),
    ],
)
def test_is_update_available(mocker, requests_mock, versions, expect):
    """Test self-update check method"""
    fake_data = {"releases": {k: [] for k in versions}}
    requests_mock.get("https://pypi.org/pypi/micropy-cli/json", json=fake_data)
    mocker.patch("micropy.__version__", "0.0.0")
    utils.helpers.get_cached_data.clear_cache()
    assert utils.helpers.is_update_available() == expect


def test_stream_download(mocker):
    """Test stream download"""
    mock_req = mocker.patch.object(utils.helpers, "requests")
    mock_stream = mocker.MagicMock()
    mock_stream.headers = {"content-length": "1000"}
    mock_req.get.return_value = mock_stream
    tqdm_mock = mocker.patch.object(utils.helpers, "tqdm")
    utils.stream_download("https://someurl.com/file.ext")
    expect_args = {
        "unit_scale": True,
        "unit_divisor": 1024,
        "smoothing": 0.1,
        "bar_format": mocker.ANY,
    }
    tqdm_mock.assert_called_once_with(total=1000, unit="B", **expect_args)



b='{}/{}'
a='_thread'
Z=IndexError
Y=NameError
X=print
W=NotImplementedError
R=True
Q='__init__'
P=open
N='version'
M=KeyError
L=ImportError
K='.'
I='_'
J=len
G=False
H=''
F=AttributeError
E='/'
D=None
A=OSError
import sys,gc as C,uos as os
from utime import sleep_us as c
from ujson import dumps as O
__version__='1.4.3'
S=2
d=2
try:from machine import resetWDT as T
except L:
	def T():0
class Stubber:
	def __init__(B,path=D,firmware_id=D):
		G=firmware_id;D=path
		try:
			if os.uname().release=='1.13.0'and os.uname().version<'v1.13-103':raise W('MicroPython 1.13.0 cannot be stubbed')
		except F:pass
		B._report=[];B.info=_info();C.collect()
		if G:B._fwid=str(G).lower()
		else:B._fwid='{family}-{port}-{ver}'.format(**B.info).lower()
		B._start_free=C.mem_free()
		if D:
			if D.endswith(E):D=D[:-1]
		else:D=get_root()
		B.path='{}/stubs/{}'.format(D,B.flat_fwid).replace('//',E)
		try:U(D+E)
		except A:pass
		B.problematic=['upip','upysh','webrepl_setup','http_client','http_client_ssl','http_server','http_server_ssl'];B.excluded=['webrepl','_webrepl','port_diag','example_sub_led.py','example_pub_button.py'];B.modules=[]
	def get_obj_attributes(K,item_instance):
		J="Couldn't get attribute '{}' from object '{}', Err: {}";A=item_instance;B=[];E=[]
		try:
			for D in dir(A):
				try:G=getattr(A,D);B.append((D,repr(G),repr(type(G)),G))
				except F as H:E.append(J.format(D,A,H))
		except F as H:E.append(J.format(D,A,H))
		B=[A for A in B if not(A[0].startswith(I)and A[0]!=Q)];C.collect();return B,E
	def add_modules(A,modules):A.modules=sorted(set(A.modules)|set(modules))
	def create_all_stubs(A):
		C.collect()
		for B in A.modules:A.create_one_stub(B)
	def create_one_stub(D,module_name):
		B=module_name
		if B.startswith(I)and B!=a:return G
		if B in D.problematic:return G
		if B in D.excluded:return G
		F='{}/{}.py'.format(D.path,B.replace(K,E));C.collect();H=C.mem_free();X('Stub module: {:<20} to file: {:<55} mem:{:>5}'.format(B,F,H))
		try:D.create_module_stub(B,F)
		except A:return G
		C.collect();return R
	def create_module_stub(F,module_name,file_name=D):
		G=file_name;B=module_name
		if B.startswith(I)and B!=a:return
		if B in F.problematic:return
		if G is D:G=F.path+E+B.replace(K,I)+'.py'
		if E in B:B=B.replace(E,K)
		J=D
		try:J=__import__(B,D,D,'*')
		except L:return
		U(G)
		with P(G,'w')as N:O='"""\nModule: \'{0}\' on {1}\n"""\n# MCU: {2}\n# Stubber: {3}\n'.format(B,F._fwid,F.info,__version__);N.write(O);N.write('from typing import Any\n\n');F.write_object_stub(N,J,B,H)
		F._report.append({'module':B,'file':G})
		if not B in['os','sys','logging','gc']:
			try:del J
			except (A,M):pass
			try:del sys.modules[B]
			except M:pass
		C.collect()
	def write_object_stub(L,fp,object_expr,obj_name,indent,in_class=0):
		b='tuple';a='list';Z='dict';X='{0}{1} = {2} # type: {3}\n';W='bound_method';V='Any';P=in_class;O=object_expr;I=fp;B=indent;C.collect()
		if O in L.problematic:return
		R,e=L.get_obj_attributes(O)
		for (D,K,F,S) in R:
			if D in['classmethod','staticmethod']:continue
			T();c(1)
			if F=="<class 'type'>"and J(B)<=d*4:E='\n'+B+'class '+D+':\n';E+=B+"    ''\n";I.write(E);L.write_object_stub(I,S,'{0}.{1}'.format(obj_name,D),B+'    ',P+1)
			elif'method'in F or'function'in F or D==Q:
				N=V;U=H
				if P>0:
					U='self, '
					if D==Q:N='None'
				if W in F or W in K:E='{}@classmethod\n'.format(B);E+='{}def {}(cls, *args) -> {}:\n'.format(B,D,N)
				else:E='{}def {}({}*args) -> {}:\n'.format(B,D,U,N)
				E+=B+'    ...\n\n';I.write(E)
			elif F=="<class 'module'>":0
			elif F.startswith("<class '"):
				G=F[8:-2];E=H
				if G in['str','int','float','bool','bytearray','bytes']:E=X.format(B,D,K,G)
				elif G in[Z,a,b]:f={Z:'{}',a:'[]',b:'()'};E=X.format(B,D,f[G],G)
				else:
					if not G in['object','set','frozenset']:G=V
					E='{0}{1} : {2} ## {3} = {4}\n'.format(B,D,G,F,K)
				I.write(E)
			else:I.write("# all other, type = '{0}'\n".format(F));I.write(B+D+' # type: Any\n')
		del R;del e
		try:del D,K,F,S
		except (A,M,Y):pass
	@property
	def flat_fwid(self):
		A=self._fwid;B=' .()/\\:$'
		for C in B:A=A.replace(C,I)
		return A
	def clean(E,path=D):
		B=path
		if B is D:B=E.path
		X('Clean/remove files in folder: {}'.format(B))
		try:G=os.listdir(B)
		except (A,F):return
		for H in G:
			C=b.format(B,H)
			try:os.remove(C)
			except A:
				try:E.clean(C);os.rmdir(C)
				except A:pass
	def report(D,filename='modules.json'):
		E=',\n';H=b.format(D.path,filename);C.collect()
		try:
			with P(H,'w')as B:
				B.write('{');B.write(O({'firmware':D.info})[1:-1]);B.write(E);B.write(O({'stubber':{N:__version__}})[1:-1]);B.write(E);B.write('"modules" :[\n');F=R
				for I in D._report:
					if F:F=G
					else:B.write(E)
					B.write(O(I))
				B.write('\n]}')
			J=D._start_free-C.mem_free()
		except A:pass
def U(path):
	C=path;B=F=0
	while B!=-1:
		B=C.find(E,F)
		if B!=-1:
			if B==0:D=C[0]
			else:D=C[0:B]
			try:I=os.stat(D)
			except A as G:
				if G.args[0]==S:
					try:os.mkdir(D)
					except A as H:raise H
				else:raise G
		F=B+1
def _info():
	g='loboris';f=' on ';e='0.0.0';d='port';c='platform';b='machine';a='nodename';Y='name';T='v';S='mpy';R='unknown';Q='-';P='sysname';I='ver';G='family';E='build';B='release';U=sys.implementation.name;V=sys.platform;A={Y:U,B:e,N:e,E:H,P:R,a:R,b:R,G:U,c:V,d:V,I:H}
	try:A[B]=K.join([str(A)for A in sys.implementation.version]);A[N]=A[B];A[Y]=sys.implementation.name;A[S]=sys.implementation.mpy
	except F:pass
	if sys.platform not in('unix','win32'):
		try:
			C=os.uname();A[P]=C.sysname;A[a]=C.nodename;A[B]=C.release;A[b]=C.machine
			if f in C.version:
				O=C.version.split(f)[0]
				if A[P]=='esp8266':
					if Q in O:W=O.split(Q)[0]
					else:W=O
					A[N]=A[B]=W.lstrip(T)
				try:A[E]=O.split(Q)[1]
				except Z:pass
		except (Z,F,TypeError):pass
	try:from pycopy import const;A[G]='pycopy';del const
	except (L,M):pass
	if A[c]=='esp32_LoBo':A[G]=g;A[d]='esp32'
	elif A[P]=='ev3':
		A[G]='ev3-pybricks';A[B]='1.0.0'
		try:from pybricks.hubs import EV3Brick;A[B]='2.0.0'
		except L:pass
	if A[B]:A[I]=T+A[B].lstrip(T)
	if A[G]!=g:
		if A[B]and A[B]>='1.10.0'and A[B].endswith('.0'):A[I]=A[B][:-2]
		else:A[I]=A[B]
		if A[E]!=H and J(A[E])<4:A[I]+=Q+A[E]
	if S in A:
		h=int(A[S]);X=[D,'x86','x64','armv6','armv6m','armv7m','armv7em','armv7emsp','armv7emdp','xtensa','xtensawin'][h>>10]
		if X:A['arch']=X
	return A
def get_root():
	B='/flash'
	try:D=os.stat(B)
	except A as C:
		if C.args[0]==S:
			try:B=os.getcwd()
			except (A,F):B=K
		else:B=E
	return B
def B():sys.exit(1)
def read_path():
	A=H
	if J(sys.argv)==3:
		C=sys.argv[1].lower()
		if C in('--path','-p'):A=sys.argv[2]
		else:B()
	elif J(sys.argv)==2:B()
	return A
def V():
	try:A=bytes('abc',encoding='utf8');B=V.__module__;return G
	except (W,F):return R
def main():
	stubber=Stubber(path=read_path());stubber.clean()
	try:stubber.modules=[A.strip()for A in P('modulelist'+'.txt')if J(A.strip())and A.strip()[0]!='#']
	except A:stubber.modules=['micropython']
	C.collect();stubber.create_all_stubs();stubber.report()
if __name__=='__main__'or V():
	try:logging.basicConfig(level=logging.INFO)
	except Y:pass
	main()


# -*- coding: utf-8 -*-

from .config_source import ConfigSource


class DictConfigSource(ConfigSource):
    def __init__(self, config: dict = {}):
        """Dict Config Source.

        Args:
            config (dict, optional): Initial Config.
                Defaults to {}.

        """
        super().__init__(initial_config=config)

    @property
    def exists(self) -> bool:
        return any(self.config.keys())

    def process(self) -> dict:
        return self.config

    def prepare(self):
        return super().prepare()

    def save(self, content: dict) -> dict:
        return content



# -*- coding: utf-8 -*-

import shutil
from pathlib import Path

import pytest
from micropy import exceptions, stubs


@pytest.fixture
def mock_fware(mocker, shared_datadir):
    def mock_ready(self, *args, **kwargs):
        fware_stub = shared_datadir / "fware_test_stub"
        return super().ready(path=fware_stub)

    fware_stub = shared_datadir / "fware_test_stub"
    mock_remote = mocker.patch.object(stubs.source, "RemoteStubSource").return_value
    mock_remote.ready.return_value.__enter__.return_value = fware_stub


def test_stub_validation(shared_datadir):
    """should pass validation"""
    stub_path = shared_datadir / "esp8266_test_stub"
    manager = stubs.StubManager()
    manager.validate(stub_path)
    assert manager.is_valid(stub_path)
    assert not manager.is_valid(Path("/foobar/bar"))


def test_bad_stub_validation(shared_datadir, mocker):
    """should fail validation"""
    stub_path = shared_datadir / "esp8266_test_stub"
    manager = stubs.StubManager()
    mock_validate = mocker.patch.object(stubs.stubs.utils, "Validator")
    mock_validate.return_value.validate.side_effect = [Exception, FileNotFoundError]
    with pytest.raises(exceptions.StubValidationError):
        manager.validate(stub_path)
    with pytest.raises(exceptions.StubError):
        manager.validate(Path("/foobar/foo"))


def test_bad_stub(tmp_path):
    """should raise exception on invalid stub"""
    with pytest.raises(FileNotFoundError):
        stubs.stubs.DeviceStub(tmp_path)


def test_valid_stub(shared_datadir):
    """should have all attributes"""
    stub_path = shared_datadir / "esp8266_test_stub"
    stub = stubs.stubs.DeviceStub(stub_path)
    stub_2 = stubs.stubs.DeviceStub(stub_path)
    fware = stubs.stubs.FirmwareStub((shared_datadir / "fware_test_stub"))
    assert stub == stub_2
    expect_fware = {
        "machine": "ESP module with ESP8266",
        "firmware": "esp8266 v1.9.4",
        "nodename": "esp8266",
        "version": "1.9.4",
        "release": "2.2.0-dev(9422289)",
        "sysname": "esp8266",
        "name": "micropython",
    }
    expect_repr = (
        "DeviceStub(sysname=esp8266, firmware=micropython," f" version=1.9.4, path={stub_path})"
    )
    assert stub.path.exists()
    assert stub.stubs.exists()
    assert stub.frozen.exists()
    assert stub.version == "1.9.4"
    assert stub.firm_info == expect_fware
    assert repr(stub) == expect_repr
    assert str(stub) == "esp8266-1.9.4"
    assert stub.firmware_name == "micropython"
    del stub.firm_info["name"]
    assert stub.firmware_name == "esp8266 v1.9.4"
    stub.firmware = fware
    assert stub.firmware_name == "micropython"
    assert str(stub) == "esp8266-micropython-1.9.4"


def test_valid_fware_stub(shared_datadir):
    stub_path = shared_datadir / "fware_test_stub"
    stub = stubs.stubs.FirmwareStub(stub_path)
    assert str(stub) == "micropython"
    assert stub.frozen.exists()
    assert repr(stub) == ("FirmwareStub(firmware=micropython," " repo=micropython/micropython)")


def test_resolve_stub(shared_datadir):
    """should resolve correct stub type"""
    device_stub = shared_datadir / "esp8266_test_stub"
    fware_stub = shared_datadir / "fware_test_stub"
    invalid_stub = shared_datadir / "esp8266_invalid_stub"
    manager = stubs.StubManager()
    stub_type = manager._get_stubtype(device_stub)
    assert stub_type == stubs.stubs.DeviceStub
    stub_type = manager._get_stubtype(fware_stub)
    assert stub_type == stubs.stubs.FirmwareStub
    with pytest.raises(exceptions.StubError):
        manager._get_stubtype(Path("/foobar/foo"))
    with pytest.raises(exceptions.StubValidationError):
        manager._get_stubtype(invalid_stub)


def test_resolve_firmware(tmp_path, shared_datadir):
    """should resolve firmware"""
    device_stub = shared_datadir / "esp8266_test_stub"
    fware_stub_path = shared_datadir / "fware_test_stub"
    manager = stubs.StubManager(resource=tmp_path)
    fware_stub = manager.add(fware_stub_path)
    dev_stub = stubs.stubs.DeviceStub(device_stub)
    resolved = manager.resolve_firmware(dev_stub)
    assert fware_stub == resolved


def test_add_single_stub(shared_datadir, tmp_path):
    """should add a single stub"""
    stub_path = shared_datadir / "esp8266_test_stub"
    manager = stubs.StubManager()
    manager.add(stub_path, dest=tmp_path)
    assert len(manager) == 1
    assert stub_path.name in [d.name for d in tmp_path.iterdir()]


def test_add_stubs_from_dir(datadir, tmp_path):
    """should add all valid stubs in directory"""
    manager = stubs.StubManager()
    manager.add(datadir, dest=tmp_path)
    assert len(manager) == 2
    assert len(list(tmp_path.iterdir())) - 1 == len(manager)
    assert manager._should_recurse(datadir)
    with pytest.raises(exceptions.StubError):
        empty_path = tmp_path / "empty"
        empty_path.mkdir()
        manager._should_recurse(empty_path)


def test_add_with_resource(datadir, mock_fware, tmp_path, mocker):
    """should not require dest kwarg"""
    resource = tmp_path / "tmp_resource"
    resource.mkdir()
    load_spy = mocker.spy(stubs.StubManager, "_load")
    manager = stubs.StubManager(resource=resource)
    manager.add(datadir)
    assert len(manager) == 2
    assert "esp8266_test_stub" in [p.name for p in resource.iterdir()]
    assert load_spy.call_count == 4
    # Should not add any new stubs
    assert manager.add(datadir)
    assert load_spy.call_count == 4
    # Should force load
    assert manager.add(datadir, force=True)
    assert load_spy.call_count == 7


def test_add_no_resource_no_dest(datadir, mock_fware):
    """should fail with typeerror"""
    manager = stubs.StubManager()
    with pytest.raises(TypeError):
        manager.add(datadir)


def test_loads_from_resource(datadir, mock_fware):
    """should load from resource if provided"""
    manager = stubs.StubManager(resource=datadir)
    assert len(manager) == 2


def test_name_property(shared_datadir):
    """should raise error if name is not overriden"""
    test_stub = shared_datadir / "esp8266_test_stub"

    class ErrorStub(stubs.stubs.Stub):
        def __init__(self, path, copy_to=None, **kwargs):
            return super().__init__(path, copy_to=copy_to, **kwargs)

    with pytest.raises(NotImplementedError):
        x = ErrorStub(test_stub)
        x.name


def test_stub_search(mocker, test_urls, shared_datadir, tmp_path, test_repo):
    test_fware = shared_datadir / "fware_test_stub"
    test_stub = shared_datadir / "esp8266_test_stub"
    tmp_path = tmp_path / "foobar"
    tmp_path.mkdir()
    manager = stubs.StubManager(resource=tmp_path, repos=[test_repo])
    manager.add(test_fware)
    print(manager._firmware)
    print(list(manager))
    manager.add(test_stub)
    results = manager.search_remote("esp8266")
    assert len(results) == 1
    res = results[0]
    assert res[0] == "esp8266-micropython-1.9.4"
    assert res[1]
    results = manager.search_remote("esp32")
    res = results[0]
    assert res[0] == "esp32-micropython-1.11.0"
    assert not res[1]


def test_stub_resolve_link(mock_mp_stubs, tmp_path):
    """should create DeviceStub from symlink"""
    stub = list(mock_mp_stubs.stubs)[0]
    link_path = tmp_path / "stub_symlink"
    linked_stub = stubs.stubs.DeviceStub.resolve_link(stub, link_path)
    assert stub == linked_stub
    assert stub.path != linked_stub.path
    assert linked_stub.path.is_symlink()
    assert linked_stub.path.resolve() == stub.path


def test_manager_resolve_subresource(mock_mp_stubs, tmp_path):
    """should create StubManager from subresource symlinks"""
    test_stubs = list(mock_mp_stubs.stubs)[:2]
    subresource = tmp_path / "stub_subresource"
    subresource.mkdir()
    manager = mock_mp_stubs.stubs.resolve_subresource(test_stubs, subresource)
    linked_stub = list(manager)[0]
    assert linked_stub.path.is_symlink()
    assert linked_stub in list(mock_mp_stubs.stubs)


def test_load_firmware_first(mocker, tmp_path, shared_datadir):
    """should always load firmware first"""
    mock_manager = mocker.patch.object(stubs.StubManager, "_load")
    mock_iterdir = mocker.patch.object(stubs.stubs.Path, "iterdir")
    # mock_mgr = mock_manager.return_value
    tmp_path = tmp_path / "fware_first_test"
    tmp_resource = tmp_path / "tmp_resource"
    tmp_resource.mkdir(parents=True)
    test_stub = shared_datadir / "esp32_test_stub"
    test_fware = shared_datadir / "fware_test_stub"
    # Ensure Firmware loads first, regardless of how Path.iterdir() orders it
    shutil.copytree(test_stub, (tmp_path / "99_esp32_test_stub"))
    shutil.copytree(test_fware, (tmp_path / "00_fware_test_stub"))
    mock_iterdir.return_value = (test_stub, test_fware)
    manager = stubs.StubManager(resource=tmp_resource)
    manager.load_from(tmp_path)
    # Get First call args
    fargs, _ = mock_manager.call_args_list[0]
    assert fargs[0].location == test_fware


def test_iter_by_firm_stubs(mocker):
    """should iter stubs by firmware"""
    firm_stub = mocker.MagicMock()
    dev_stub = mocker.MagicMock()
    dev_stub.firmware = firm_stub
    unk_stub = mocker.MagicMock()
    unk_stub.firmware = None
    manager = stubs.StubManager()
    manager._loaded = set([firm_stub, dev_stub, unk_stub])
    manager._firmware = set([firm_stub])
    stub_iter = list(manager.iter_by_firmware())
    assert stub_iter == [(firm_stub, [dev_stub]), ("Unknown", [unk_stub])]



# -*- coding: utf-8 -*-

"""
micropy.utils.decorators
~~~~~~~~~~~~~~

This module contains generic decorators
used by MicropyCli
"""

__all__ = ["lazy_property"]


def lazy_property(fn):
    attr = "_lazy__" + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr):
            setattr(self, attr, fn(self))
        return getattr(self, attr)

    return _lazy_property



# -*- coding: utf-8 -*-

"""Logging functionality.

TODO: Split logging from UI, refactor.

"""

import logging
import re
from contextlib import contextmanager
from logging.handlers import RotatingFileHandler

import click
from micropy import data


class Log:
    """Borg for easy access to any Log from anywhere in the package."""

    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state
        self.parent_logger = ServiceLog()
        self.loggers = [self.parent_logger]

    @classmethod
    def add_logger(cls, service_name, base_color="white", **kwargs):
        """Creates a new child ServiceLog instance."""
        _self = cls()
        parent = kwargs.pop("parent", _self.parent_logger)
        logger = ServiceLog(service_name, base_color, parent=parent, **kwargs)
        _self.loggers.append(logger)
        return logger

    @classmethod
    def get_logger(cls, service_name):
        """Retrieves a child logger by service name."""
        _self = cls()
        logger = next((i for i in _self.loggers if i.service_name == service_name))
        return logger


class ServiceLog:
    """Handles logging to stdout and micropy.log.

    :param service_name: Active service to display
    :type service_name: str
    :param base_color: Color of name on output
    :type base_color: str

    """

    LOG_FILE = data.LOG_FILE

    def __init__(self, service_name="MicroPy", base_color="bright_green", **kwargs):
        self.parent = kwargs.get("parent", None)
        self.LOG_FILE.parent.mkdir(exist_ok=True)
        self.base_color = base_color
        self.service_name = service_name
        self.load_handler()
        self.info_color = kwargs.get("info_color", "white")
        self.accent_color = kwargs.get("accent_color", "yellow")
        self.warn_color = kwargs.get("warn_color", "green")
        self.show_title = kwargs.get("show_title", True)
        self.stdout = kwargs.get("stdout", True)

    @contextmanager
    def silent(self):
        self.stdout = False
        yield self
        self.stdout = True

    def load_handler(self):
        """Loads Logging Module Formatting."""
        self.log = logging.getLogger()
        if not self.log.hasHandlers():
            self.log.setLevel(logging.DEBUG)
            self.log_handler = RotatingFileHandler(
                str(self.LOG_FILE),
                mode="a",
                maxBytes=2 * 1024 * 1024,
                backupCount=2,
                encoding=None,
                delay=0,
            )
            self.log_handler.setLevel(logging.DEBUG)
            self.log.addHandler(self.log_handler)
        self.log_handler = self.log.handlers[0]
        log_format = "[%(asctime)s] %(levelname)s: " f"{self.service_name.lower()}: " "%(message)s"
        self.log_handler.setFormatter(logging.Formatter(log_format, "%Y-%m-%d %H:%M:%S"))

    def parse_msg(self, msg, accent_color=None):
        """Parses any color codes accordingly.

        :param str msg:
        :param str accent_color:  (Default value = None)
        :return: Parsed Message
        :rtype: str

        """
        msg_special = re.findall(r"\$(.*?)\[(.*?)\]", msg)
        color = accent_color or self.accent_color
        special = {"fg": color, "bold": True}
        clean = msg
        _parts = re.split(r"\$.*?\[(.*?)\]", msg)
        parts = [(p, None) for p in _parts]
        for w in msg_special:
            if w[0] == "w":
                special["fg"] = self.warn_color
            if w[0] == "B":
                special.pop("fg")
            sindex = _parts.index(w[1])
            parts[sindex] = (w[1], special)
            clean = msg.replace(f"${w[0]}[{w[1]}]", w[1])
        clean = clean.encode("ascii", "ignore").decode("utf-8").strip()
        return (parts, clean)

    def get_parents(self, names=[]):
        """Retrieve all parents."""
        if len(names) == 0:
            names = [self.service_name]
        if self.parent:
            names.insert(0, self.parent.service_name)
            names = self.parent.get_parents(names)
        return names

    def get_service(self, **kwargs):
        """Retrieves formatted service title.

        :param **kwargs:
        :return: formatted title
        :rtype: str

        """
        if not self.show_title:
            return f"{self.parent.get_service(bold=True)}"
        color = kwargs.pop("fg", self.base_color)
        title = click.style(f"{self.service_name}", fg=color, **kwargs)
        title = f"{title}{click.style(' ', fg=color)}"
        if self.parent is not None:
            title = f"{self.parent.get_service(bold=True)} {title}"
        return title

    def iter_formatted(self, message, **kwargs):
        """Iterate formatted message tuple into styled string.

        Args:
            message (tuple): tuple as (msg, style)

        """
        if isinstance(message, str):
            message, _ = self.parse_msg(message)
        for msg in message:
            text, mstyle = msg
            mstyle = mstyle or kwargs
            yield click.style(text, **mstyle)

    def echo(self, msg, **kwargs):
        """Prints msg to stdout.

        :param str msg: message to print
        :param **kwargs:

        """
        title_color = kwargs.pop("title_color", self.base_color)
        title_bold = kwargs.pop("title_bold", True)
        accent_color = kwargs.pop("accent", self.accent_color)
        service_title = self.get_service(fg=title_color, bold=title_bold)
        message, clean = self.parse_msg(msg, accent_color)
        log_attr = kwargs.pop("log", None)
        if log_attr:
            self.load_handler()
            log_func = getattr(logging, log_attr)
            log_func(clean)
        if self.stdout:
            init_msg, init_style = message[0]
            first_part, nl_part, _ = init_msg.partition("\n")
            fp_clean = first_part.encode("ascii", "ignore").decode("unicode_escape")
            if not fp_clean.strip() and nl_part == "\n":
                init_msg = init_msg.replace("\n", "")
                message[0] = (init_msg, init_style)
                click.secho("")
            click.secho(f"{service_title} ", nl=False)
            post_nl = kwargs.pop("nl", None)
            formatted = list(self.iter_formatted(message, **kwargs))
            for msg in formatted:
                do_nl = msg == formatted[-1]
                click.echo(msg, nl=do_nl)
            if post_nl:
                click.echo("")

    def info(self, msg, **kwargs):
        """Prints message with info formatting.

        :param msg:
        :param **kwargs:
        :return: method to print msg
        :rtype: method

        """
        return self.echo(msg, log="info", **kwargs)

    def title(self, msg, **kwargs):
        """Prints bolded info message.

        Args:
            msg (str): Message

        """
        return self.info(f"\n{msg}", bold=True)

    def error(self, msg, exception=None, **kwargs):
        """Prints message with error formatting.

        :param msg:
        :param **kwargs:
        :return: method to print msg
        :rtype: method

        """
        bold = kwargs.pop("bold", (exception != None))
        self.echo(
            msg,
            log="error",
            title_color="red",
            title_bold=True,
            fg="red",
            accent="red",
            bold=bold,
            **kwargs,
        )
        if exception:
            return self.exception(exception)

    def warn(self, msg, **kwargs):
        """Prints message with warn formatting.

        :param msg:
        :param **kwargs:
        :return: method to print msg
        :rtype: method

        """
        return self.echo(msg, log="warning", title_color="red", title_bold=True)

    def exception(self, error, **kwargs):
        """Prints message with exception formatting.

        :param error:
        :param **kwargs:
        :return: method to print msg
        :rtype: method

        """
        name = type(error).__name__
        msg = f"{name}: {str(error)}"
        return self.echo(msg, log="exception", title_color="red", fg="red", accent="red", **kwargs)

    def success(self, msg, **kwargs):
        """Prints message with success formatting.

        :param msg:
        :param **kwargs:
        :return: method to print msg
        :rtype: method
        :return: method to print msg
        :rtype: method

        """
        message = f"\u2714 {msg}"
        return self.echo(message, log="info", fg="green", **kwargs)

    def debug(self, msg, **kwargs):
        """Prints message with debug formatting.

        :param msg:
        :param **kwargs:
        :return: method to log msg
        :rtype: method

        """
        if self.stdout:
            with self.silent():
                return self.debug(msg, **kwargs)
        self.echo(msg, log="debug")
        return msg



# -*- coding: utf-8 -*-

"""Main Module."""

import tempfile
from pathlib import Path

from micropy import data, utils
from micropy.exceptions import PyDeviceError
from micropy.lib.stubber import process as stubber
from micropy.logger import Log
from micropy.project import Project, modules
from micropy.pyd import DevicePath, MessageHandlers, ProgressStreamConsumer, PyDevice
from micropy.stubs import StubManager, source


class MicroPy:
    """Handles App State Management."""

    RUN_CHECKS = True

    def __init__(self):
        self.log = Log.get_logger("MicroPy")
        self.verbose = True
        self.log.debug("MicroPy Loaded")
        if not data.STUB_DIR.exists():
            self.setup()

    def setup(self):
        """Creates necessary directories for micropy."""
        self.log.debug("Running first time setup...")
        self.log.debug(f"Creating .micropy directory @ {data.FILES}")
        data.FILES.mkdir(exist_ok=True)
        data.STUB_DIR.mkdir()

    @utils.lazy_property
    def stubs(self):
        """Primary Stub Manager for MicroPy.

        Returns:
            StubManager: StubManager Instance

        """
        repo_list = data.REPO_SOURCES.read_text()
        repos = source.StubRepo.from_json(repo_list)
        return StubManager(resource=data.STUB_DIR, repos=repos)

    @utils.lazy_property
    def project(self):
        """Current active project if available.

        Returns:
            Project: Instance of Current Project

        """
        proj = self.resolve_project(".", verbose=self.verbose)
        return proj

    def resolve_project(self, path, verbose=True):
        """Returns project from path if it exists.

        Args:
            path (str): Path to test
            verbose (bool): Log to stdout. Defaults to True.

        Returns:
            Project if it exists

        """
        path = Path(path).absolute()
        proj = Project(path)
        proj.add(modules.StubsModule, self.stubs)
        proj.add(modules.PackagesModule, "requirements.txt")
        proj.add(modules.DevPackagesModule, "dev-requirements.txt")
        proj.add(modules.TemplatesModule, run_checks=self.RUN_CHECKS)
        if proj.exists:
            if verbose:
                self.log.title(f"Loading Project")
            proj.load()
            if verbose:
                self.log.success("Ready!")
            return proj
        return proj

    def create_stubs(self, port, verbose=False):
        """Create and add stubs from Pyboard.

        Todo:
            Extract and cleanup this mess.

        Args:
            port (str): Port of Pyboard

        Returns:
            Stub: generated stub

        """
        self.log.title(f"Connecting to Pyboard @ $[{port}]")
        pyb_log = Log.add_logger("Pyboard", "bright_white")

        def _get_desc(name: str, cfg: dict):
            desc = f"{pyb_log.get_service()} {name}"
            return name, cfg | dict(desc=desc)

        message_handler = MessageHandlers(
            on_message=lambda x: isinstance(x, str) and pyb_log.info(x.strip())
        )
        try:
            pyb = PyDevice(
                port,
                auto_connect=True,
                stream_consumer=ProgressStreamConsumer(on_description=_get_desc),
                message_consumer=message_handler,
            )
        except (SystemExit, PyDeviceError):
            self.log.error(f"Failed to connect, are you sure $[{port}] is correct?")
            return None
        self.log.success("Connected!")
        script = stubber.minify_script(stubber.source_script)
        self.log.info("Executing stubber on pyboard...")
        try:
            pyb.run_script(script, DevicePath("createstubs.py"))
        except Exception as e:
            # TODO: Handle more usage cases
            self.log.error(f"Failed to execute script: {str(e)}", exception=e)
            raise
        self.log.success("Done!")
        self.log.info("Copying stubs...")
        with tempfile.TemporaryDirectory() as tmpdir:
            pyb.copy_from(DevicePath("/stubs"), tmpdir)
            out_dir = Path(tmpdir)
            stub_path = next(out_dir.iterdir())
            self.log.info(f"Copied Stubs: $[{stub_path.name}]")
            stub_path = self.stubs.from_stubber(stub_path, out_dir)
            stub = self.stubs.add(stub_path)
        pyb.disconnect()
        self.log.success(f"Added {stub.name} to stubs!")
        return stub



from __future__ import annotations

from functools import partialmethod
from typing import Any, Callable, NamedTuple, cast

from micropy.pyd.abc import (
    EndHandler,
    MessageConsumer,
    MessageHandler,
    StartHandler,
    StreamConsumer,
    UpdateHandler,
)
from tqdm import tqdm


class ProgressStreamConsumer:
    bar: tqdm

    def __init__(
        self,
        on_description: Callable[
            [str, dict[str, Any] | None], tuple[str, dict[str, Any] | None]
        ] = None,
        **kwargs,
    ):
        self._on_description = on_description or (lambda s, cfg: (s, cfg))

    def on_start(self, *, name: str = None, size: int | None = None):
        bar_format = "{l_bar}{bar}| [{n_fmt}/{total_fmt} @ {rate_fmt}]"
        tqdm_kwargs = {
            "unit_scale": True,
            "unit_divisor": 1024,
            "bar_format": bar_format,
        }
        _name, _tqdm_kws = self._on_description(name or "", tqdm_kwargs)
        # todo: use union operator when min py version is 3.9.
        tqdm_kwargs.update(_tqdm_kws or dict())
        self.bar = tqdm(total=size, unit="B", **tqdm_kwargs)

    def on_update(self, *, size: int | None = None):
        self.bar.update(size)

    def on_end(self):
        self.bar.close()


class ConsumerDelegate:
    consumers: list[StreamConsumer | MessageConsumer]

    def __init__(self, *consumers: StreamConsumer | MessageConsumer | None):
        self.consumers = [i for i in consumers if i]

    def consumer_for(self, action: str, *args, **kwargs):
        _consumer = next((i for i in self.consumers if hasattr(i, action)), None)
        if _consumer is None:
            # default noop
            return
        return getattr(_consumer, action)(*args, **kwargs)

    on_message = cast(MessageHandler, partialmethod(consumer_for, "on_message"))
    on_start = cast(StartHandler, partialmethod(consumer_for, "on_start"))
    on_update = cast(UpdateHandler, partialmethod(consumer_for, "on_update"))
    on_end = cast(EndHandler, partialmethod(consumer_for, "on_end"))


class StreamHandlers(NamedTuple):
    on_start: StartHandler
    on_update: UpdateHandler
    on_end: EndHandler


class MessageHandlers(NamedTuple):
    on_message: MessageHandler



# -*- coding: utf-8 -*-

"""Config Abstract."""

import abc
import contextlib
from typing import Any

from micropy.logger import Log, ServiceLog


class ConfigSource(contextlib.AbstractContextManager, metaclass=abc.ABCMeta):
    """Abstract Base Class for Config Sources.

    Args:
        initial_config (dict, optional): Initial config values.
            Defaults to {}.

    """

    def __init__(self, initial_config: dict = {}):
        self._config: dict = initial_config
        self.log: ServiceLog = Log.add_logger(__name__)

    @property
    def config(self) -> dict:
        """Current Config Content."""
        return self._config

    @config.setter
    def config(self, value: dict) -> dict:
        """Set current config content.

        Args:
            value (dict): New value to set

        Returns:
            dict: Current config

        """
        self._config = value
        return self._config

    @abc.abstractproperty
    def exists(self) -> bool:
        """Property to check if source exists."""

    @abc.abstractmethod
    def save(self, content: Any) -> Any:
        """Method to save config."""

    @abc.abstractmethod
    def process(self) -> dict:
        """Read and process config file.

        Returns:
            dict: Config file content

        """

    @abc.abstractmethod
    def prepare(self) -> Any:
        """Method to prepare on enter."""

    def __enter__(self) -> dict:
        self.prepare()
        self._config = self.process()
        return self._config

    def __exit__(self, *args):
        self.save(self._config)
